import binascii
import statistics
from typing import TypeVar, Union, Literal
from functools import reduce as functools_reduce

from ..core import ChepyCore, ChepyDecorators
from .exceptions import StateNotList
from .internal.helpers import detect_delimiter


AritmeticLogicT = TypeVar("AritmeticLogicT", bound="AritmeticLogic")


class AritmeticLogic(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    def __hex_to_int(self, n):  # pragma: no cover
        if isinstance(n, str):
            return int(n, 0)
        if isinstance(n, int):
            return n

    @ChepyDecorators.call_stack
    def add(self, n: int) -> AritmeticLogicT:
        """Add a number to the state

        Args:
            n (int): Number to add with. Can be decimal or hex string without 0x

        Returns:
            Chepy: The Chepy object.
        """
        # Determine the base of the key (hexadecimal or decimal)
        if isinstance(n, int):
            # Try converting to decimal
            key_int = n
        else:
            try:
                # Try converting to hexadecimal
                key_int = int(n, 16)
            except ValueError:  # pragma: no cover
                self._log.error(
                    "Invalid key format. Must be a decimal or hexadecimal string."
                )
                return self

        hold = b""
        for char_code in self._convert_to_bytes():
            # Add the key to the integer and take the result modulo 255
            result_code = (char_code + key_int) % 256

            # Convert the result back to a byte
            hold += result_code.to_bytes(1, byteorder="big")

        self.state = hold.decode()
        return self

    @ChepyDecorators.call_stack
    def addition(self, delimiter=None) -> AritmeticLogicT:
        """Adds a list of numbers. If an item in the string is not a number it is excluded from the list.

        Args:
            delimiter (str, optional): Delimiter. Defaults to None.

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_str()
        if not delimiter:
            delimiter = detect_delimiter(data)
        # only work on numbers
        nums = []
        for n in data.split(delimiter):
            try:
                nums.append(int(n))
            except:  # noqa: E722
                continue

        self.state = functools_reduce(lambda x, y: x + y, nums)
        return self

    @ChepyDecorators.call_stack
    def sub(self, n: int) -> AritmeticLogicT:
        """SUB the input with the given key

        Args:
            n (int): Number to subtract with

        Returns:
            Chepy: The Chepy object.
        """
        # Determine the base of the key (hexadecimal or decimal)
        if isinstance(n, int):
            # Try converting to decimal
            key_int = n
        else:
            try:
                # Try converting to hexadecimal
                key_int = int(n, 16)
            except ValueError:  # pragma: no cover
                self._log.error(
                    "Invalid key format. Must be a decimal or hexadecimal string."
                )
                return self

        hold = b""
        for char_code in self._convert_to_bytes():
            # Add the key to the integer and take the result modulo 255
            result_code = (char_code - key_int) % 256

            # Convert the result back to a byte
            hold += result_code.to_bytes(1, byteorder="big")

        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def subtract(self, delimiter=None) -> AritmeticLogicT:
        """Subtracts a list of numbers. If an item in the string is not a number it is excluded from the list.

        Args:
            delimiter (str, optional): Delimiter. Defaults to None.

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_str()
        if not delimiter:
            delimiter = detect_delimiter(data)
        # only work on numbers
        nums = []
        for n in data.split(delimiter):
            try:
                nums.append(int(n))
            except:  # noqa: E722
                continue

        self.state = functools_reduce(lambda x, y: x - y, nums)
        return self

    @ChepyDecorators.call_stack
    def multiply(self, n: int) -> AritmeticLogicT:
        """Multiply a number to the state

        Args:
            n (int): Number to multiply with

        Returns:
            Chepy: The Chepy object.
        """
        if not isinstance(self.state, int):
            self.state = self.__hex_to_int(self.state)
        self.state = self.state * n
        return self

    @ChepyDecorators.call_stack
    def divide(self, n: int) -> AritmeticLogicT:
        """Divide a number to the state. Chepy is not optimized for float math.
        Subsequent methods may fail.

        Args:
            n (int): Number to divide with

        Returns:
            Chepy: The Chepy object.
        """
        if not isinstance(self.state, int):
            self.state = self.__hex_to_int(self.state)
        self.state = self.state / n
        return self

    @ChepyDecorators.call_stack
    def power(self, n: int) -> AritmeticLogicT:
        """Convert state to the n power of

        Args:
            n (int): Exponent

        Returns:
            Chepy: The Chepy object.
        """
        if not isinstance(self.state, int):
            self.state = self.__hex_to_int(self.state)
        self.state = self.state**n
        return self

    @ChepyDecorators.call_stack
    def sum(self) -> AritmeticLogicT:
        """Calculate the sum of the state

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, list), StateNotList()
        numbers = list(self.__hex_to_int(x) for x in self.state)
        self.state = sum(numbers)
        return self

    @ChepyDecorators.call_stack
    def mean(self) -> AritmeticLogicT:
        """Calculate the mean of the state

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, list), StateNotList()
        numbers = list(self.__hex_to_int(x) for x in self.state)
        self.state = statistics.mean(numbers)
        return self

    @ChepyDecorators.call_stack
    def median(self) -> AritmeticLogicT:
        """Calculate the median of the state

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, list), StateNotList()
        numbers = list(self.__hex_to_int(x) for x in self.state)
        self.state = statistics.median(numbers)
        return self

    @ChepyDecorators.call_stack
    def int_to_base(self, base: Union[int, str]) -> AritmeticLogicT:
        """Convert the state to a different base

        Args:
            base (int): Base to convert to

        Returns:
            Chepy: The Chepy object.
        """
        self.state = int(self.state, base)
        return self

    @ChepyDecorators.call_stack
    def bit_shift_right(
        self,
        amount: int = 1,
        operation_type: Literal["logical", "arithmetic"] = "logical",
    ) -> AritmeticLogicT:
        """Shifts the bits in each byte towards the right by the specified amount.

        Args:
            amount (int, optional): Amount. Defaults to 1
            operation_type (Literal['logical', 'arithmetic'], optional): Operation type. Defaults to 'logical'.

        Returns:
            Chepy: The Chepy object.
        """
        mask = 0x80 if operation_type.lower() != "logical" else 0
        output_bytes = [
            (byte >> int(amount)) ^ (byte & mask) for byte in self._convert_to_bytes()
        ]
        self.state = bytearray(output_bytes)
        return self

    @ChepyDecorators.call_stack
    def bit_shift_left(self, amount: int = 1):
        """Shifts each byte in the input byte array to the left by a specified amount.

        Args:
            amount (int, optional): Amount. Defaults to 1.

        Returns:
            Chepy: The Chepy object.
        """
        output_bytes = [(byte << amount) & 0xFF for byte in self._convert_to_bytes()]
        self.state = bytearray(output_bytes)
        return self



from typing import Any, Iterator, List

def generate_combo(words: List[Any], min_length: int=..., max_length: int=..., join_by: str=...) -> Iterator[tuple]: ...
def hex_chars() -> list: ...
def generate_random_string(length: int=8) -> str: ...



import binascii
import hashlib

import lazy_import

from typing import TypeVar, Union, Any

from crccheck.crc import Crc8, Crc32, CrcArc
from typing_extensions import Literal

HashingT = TypeVar("HashingT", bound="Hashing")

hmac = lazy_import.lazy_module("hmac")
MD2 = lazy_import.lazy_module("Crypto.Hash.MD2")
MD4 = lazy_import.lazy_module("Crypto.Hash.MD4")
MD5 = lazy_import.lazy_module("Crypto.Hash.MD5")
SHA512 = lazy_import.lazy_module("Crypto.Hash.SHA512")
SHA1 = lazy_import.lazy_module("Crypto.Hash.SHA1")
SHA256 = lazy_import.lazy_module("Crypto.Hash.SHA256")
keccak = lazy_import.lazy_module("Crypto.Hash.keccak")
SHAKE128 = lazy_import.lazy_module("Crypto.Hash.SHAKE128")
SHAKE256 = lazy_import.lazy_module("Crypto.Hash.SHAKE256")
RIPEMD = lazy_import.lazy_module("Crypto.Hash.RIPEMD")
BLAKE2s = lazy_import.lazy_module("Crypto.Hash.BLAKE2s")
BLAKE2b = lazy_import.lazy_module("Crypto.Hash.BLAKE2b")
import passlib.hash

# from Crypto.Protocol.KDF import PBKDF2
KDF = lazy_import.lazy_module("Crypto.Protocol.KDF")
# from Crypto.Protocol.KDF import bcrypt as _crypto_bcrypt
# from Crypto.Protocol.KDF import bcrypt_check as _crypto_bcrypt_check
# from Crypto.Protocol.KDF import scrypt as _crypto_scrypt

from ..core import ChepyCore, ChepyDecorators


class Hashing(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    @ChepyDecorators.call_stack
    def sha1(self) -> HashingT:
        """Get SHA1 hash

        The SHA (Secure Hash Algorithm) hash functions were designed by the NSA.
        SHA-1 is the most established of the existing SHA hash functions and it is
        used in a variety of security applications and protocols. However, SHA-1's
        collision resistance has been weakening as new attacks are discovered or improved.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("A").sha1().out
            "6dcd4ce23d88e2ee9568ba546c007c63d9131c1b"
        """
        self.state = hashlib.sha1(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha2_256(self) -> HashingT:
        """Get SHA2-256 hash

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("A").sha2_256().out
            "559aead08264d5795d3909718cdd05abd49572e84fe55590eef31a88a08fdffd"
        """
        self.state = hashlib.sha256(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha2_512(self) -> HashingT:
        """Get SHA2-512 hash

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("A").sha2_512().out
            21b4f4bd9e64ed355c3eb676a28ebedaf6d8f17bdc365995b319097153044080516bd083bfcce66121a3072646994c8430cc382b8dc543e84880183bf856cff5
        """
        self.state = hashlib.sha512(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha2_512_truncate(self, truncate: int = 256) -> HashingT:
        """Get SHA2-512/bits hash

        Args:
            truncate (int, optional): The bits to truncate by. Defaults to 256

        Returns:
            Chepy: The Chepy object.
        """
        assert truncate in [256, 224], "Valid truncates are 256, 224"
        h = SHA512.new(self._convert_to_bytes(), truncate=str(truncate))
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha2_384(self) -> HashingT:
        """Get SHA2-384 hash

        Returns:
            Chepy: The Chepy object.
        """
        self.state = hashlib.sha384(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha2_224(self) -> HashingT:
        """Get SHA2-224 hash

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("A").sha2_224().out
            "5cfe2cddbb9940fb4d8505e25ea77e763a0077693dbb01b1a6aa94f2"
        """
        self.state = hashlib.sha224(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha3_512(self) -> HashingT:
        """Get SHA3-512 hash

        Returns:
            Chepy: The Chepy object.
        """
        self.state = hashlib.sha3_512(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha3_256(self) -> HashingT:
        """Get SHA3-256 hash

        Returns:
            Chepy: The Chepy object.
        """
        self.state = hashlib.sha3_256(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha3_384(self) -> HashingT:
        """Get SHA3-384 hash

        Returns:
            Chepy: The Chepy object.
        """
        self.state = hashlib.sha3_384(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def sha3_224(self) -> HashingT:
        """Get SHA2-224 hash

        Returns:
            Chepy: The Chepy object.
        """
        self.state = hashlib.sha3_224(self._convert_to_bytes()).hexdigest()
        return self

    @ChepyDecorators.call_stack
    def md2(self) -> HashingT:
        """Get MD2 hash

        Returns:
            Chepy: The Chepy object.
        """
        h = MD2.new()
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def md4(self) -> HashingT:
        """Get MD4 hash

        Returns:
            Chepy: The Chepy object.
        """
        h = MD4.new()
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def md5(self) -> HashingT:
        """Get MD5 hash

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("A").md5().out
            "7fc56270e7a70fa81a5935b72eacbe29"
        """
        h = MD5.new()
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def keccak_512(self) -> HashingT:
        """Get KECCAK-512 hash

        Returns:
            Chepy: The Chepy object.
        """
        h = keccak.new(digest_bits=512)
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def keccak_384(self) -> HashingT:
        """Get KECCAK-384 hash

        Returns:
            Chepy: The Chepy object.
        """
        h = keccak.new(digest_bits=384)
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def keccak_256(self) -> HashingT:
        """Get KECCAK-256 hash

        Returns:
            Chepy: The Chepy object.
        """
        h = keccak.new(digest_bits=256)
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def keccak_224(self) -> HashingT:
        """Get KECCAK-224 hash

        Returns:
            Chepy: The Chepy object.
        """
        h = keccak.new(digest_bits=224)
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def shake_256(self, size: int = 64) -> HashingT:
        """Get Shake-256 hash

        Shake is an Extendable Output Function (XOF) of the SHA-3 hash algorithm,
        part of the Keccak family, allowing for variable output length/size.

        Args:
            size (int, optional): How many bytes to read, by default 64

        Returns:
            Chepy: The Chepy object.
        """
        h = SHAKE256.new()
        h.update(self._convert_to_bytes())
        self.state = binascii.hexlify(h.read(size))
        return self

    @ChepyDecorators.call_stack
    def shake_128(self, size: int = 64) -> HashingT:
        """Get Shake-128 hash

        Shake is an Extendable Output Function (XOF) of the SHA-3 hash algorithm,
        part of the Keccak family, allowing for variable output length/size.

        Args:
            size (int, optional): How many bytes to read, by default 64

        Returns:
            Chepy: The Chepy object.
        """
        h = SHAKE128.new()
        h.update(self._convert_to_bytes())
        self.state = binascii.hexlify(h.read(size))
        return self

    @ChepyDecorators.call_stack
    def ripemd_160(self) -> HashingT:
        """Get RIPEMD-160 hash

        Returns:
            Chepy: The Chepy object.
        """
        h = RIPEMD.new()
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def blake_2b(self, bits: int = 256, key: bytes = "") -> HashingT:
        """Get Balke-2b hash

        Args:
            bits (int, optional): Number of digest bits, by default 256
            key (bytes, optional): Encryption secret key, by default ''

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("A").blake_2b(bits=128, key="key").out
            "6d2e4cba3bc564e02d1a76f585a6795d"
        """
        assert bits in [
            512,
            384,
            256,
            160,
            128,
        ], "Valid bits are 512, 384, 256, 160, 128"
        h = BLAKE2b.new(digest_bits=bits, key=key.encode())
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def blake_2s(self, bits: int = 256, key: bytes = "") -> HashingT:
        """Get Blake-2s hash

        Args:
            bits (int, optional): Number of digest bits, by default 256
            key (bytes, optional): Encryption secret key, by default ''

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("A").blake_2s(bits=128, key="key").out
            "4e33cc702e9d08c28a5e9691f23bc66a"
        """
        assert bits in [256, 160, 128], "Valid bits are 256, 160, 128"
        h = BLAKE2s.new(digest_bits=bits, key=key.encode())
        h.update(self._convert_to_bytes())
        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def crc8_checksum(self) -> HashingT:
        """Get CRC8 checksum

        Returns:
            Chepy: The Chepy object.
        """
        self.state = Crc8().process(self._convert_to_bytes()).finalhex()
        return self

    @ChepyDecorators.call_stack
    def crc16_checksum(self) -> HashingT:
        """Get CRC16 checksum

        Returns:
            Chepy: The Chepy object.
        """
        self.state = CrcArc().process(self._convert_to_bytes()).finalhex()
        return self

    @ChepyDecorators.call_stack
    def crc32_checksum(self) -> HashingT:
        """Get CRC32 checksum

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("a").crc32_checksum().out
            "e8b7be43"
        """
        self.state = Crc32().process(self._convert_to_bytes()).finalhex()
        return self

    @ChepyDecorators.call_stack
    def hmac_hash(self, key: bytes = b"", digest: str = "sha1") -> HashingT:
        """Get HMAC hash

        HMAC hash the state

        Keyed-Hash Message Authentication Codes (HMAC) are a mechanism for
        message authentication using cryptographic hash functions.

        Args:
            key (bytes, optional): Starting key for the hash, by default b''
            digest (str, optional): The digest type, by default "sha1". Possible values are
                md5, sha1, sha256 and sha512

        Returns:
            Chepy: The Chepy object.

        Raises:
            TypeError: If key is not in bytes
            TypeError: If not a valid/allowed digest type
        """
        if not isinstance(key, bytes):
            if isinstance(key, str):
                key = key.encode()
            elif isinstance(key, int):  # pragma: no cover
                key = bytes(key)
            else:  # pragma: no cover
                raise TypeError("key has to be bytes")

        if digest == "md5":
            h = hmac.new(key, self._convert_to_bytes(), hashlib.md5)
        elif digest == "sha1":
            h = hmac.new(key, self._convert_to_bytes(), hashlib.sha1)
        elif digest == "sha256":
            h = hmac.new(key, self._convert_to_bytes(), hashlib.sha256)
        elif digest == "sha512":
            h = hmac.new(key, self._convert_to_bytes(), hashlib.sha512)
        else:  # pragma: no cover
            raise TypeError(
                "Currently supported digests are md5, sha1, sha256 and sha512"
            )

        self.state = h.hexdigest()
        return self

    @ChepyDecorators.call_stack
    def derive_pbkdf2_key(
        self,
        password: Union[str, bytes],
        salt: Union[str, bytes],
        key_size: int = 256,
        iterations: int = 1000,
        hash_type: Literal["md5", "sha1", "sha256", "sha512"] = "sha1",
        hex_salt: bool = True,
        show_full_key: bool = False,
    ):
        """Derive a PBKDF2 key

        Args:
            password (Union[str, bytes]): Password
            salt (Union[str, bytes]): Salt
            key_size (int, optional): Key size. Defaults to 256.
            iterations (int, optional): Number of iterations. Defaults to 1000.
            hash_type (Literal[, optional): Hash type. Defaults to "sha1".
            hex_salt (bool, optional): If salt is in hex format. Defaults to True.
            show_full_key (bool, optional): Show AES256 64 bit key only. Defaults to False.

        Raises:
            TypeError: If hash type is not one of md5, sha1, sha256, sha512

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(".").derive_pbkdf2_key("mR3m", "d9016d44c374f5fb62604683f4d61578").o[:10]
            "7c8898f222"
        """
        if isinstance(salt, str):
            salt = salt.encode()

        if hex_salt:
            salt = binascii.unhexlify(salt)

        if hash_type == "md5":
            h = KDF.PBKDF2(
                password, salt, key_size, count=iterations, hmac_hash_module=MD5
            )
        elif hash_type == "sha1":
            h = KDF.PBKDF2(
                password, salt, key_size, count=iterations, hmac_hash_module=SHA1
            )
        elif hash_type == "sha256":
            h = KDF.PBKDF2(
                password, salt, key_size, count=iterations, hmac_hash_module=SHA256
            )
        elif hash_type == "sha512":
            h = KDF.PBKDF2(
                password, salt, key_size, count=iterations, hmac_hash_module=SHA512
            )
        else:  # pragma: no cover
            raise TypeError(
                "Currently supported digests are md5, sha1, sha256 and sha512"
            )

        if show_full_key:
            self.state = h.hex()
        else:
            self.state = h.hex()[:64]
        return self

    @ChepyDecorators.call_stack
    def bcrypt_hash(self, rounds: int = 10) -> HashingT:
        """Get Bcrypt hash

        Args:
            rounds (int, optional): rounds of hashing, by default 10

        Returns:
            Chepy: The Chepy object.
        """
        self.state = KDF.bcrypt(self._convert_to_str(), cost=rounds)
        return self

    @ChepyDecorators.call_stack
    def bcrypt_compare(self, hash: str) -> HashingT:
        """Compare Bcrypt hash

        Tests whether the provided hash matches the given string at init.

        Args:
            hash (str): Required. brypt hash

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("abc")
            >>> c.bcrypt_compare("$2a$10$SpXMRnrQ4IQqC710xMHfAu0BBr4nJkuPqDvzhiAACnykgn87iE2S2")
            True
        """
        try:
            if KDF.bcrypt_check(self._convert_to_str(), hash) is None:
                self.state = True
                return self
            else:  # pragma: no cover
                self.state = False
                return self
        except ValueError:  # pragma: no cover
            self.state = False
            return self

    @ChepyDecorators.call_stack
    def scrypt_hash(
        self, salt: str = "", key_length: int = 64, N: int = 14, r: int = 8, p: int = 1
    ):
        """Get Scrypt hash

        Args:
            salt (str, optional): A string of characters that modifies the hash. Defaults to "".
            key_length (int, optional): number of bytes to use when autogenerating new salts. Defaults to 64.
            N (int, optional): CPU/memory cost parameter. Defaults to 14.
            r (int, optional): The blocksize parameter. Defaults to 8.
            p (int, optional): Parallelization parameter;. Defaults to 1.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("abc").scrypt_hash(salt="", key_length=16).out
            "f352f3374cf4e344dde4108b96985248"
        """
        assert N < 32, "N must be less than 32"
        self.state = KDF.scrypt(
            self._convert_to_bytes(), salt=salt, key_len=key_length, N=2**N, r=r, p=p
        ).hex()
        return self

    @ChepyDecorators.call_stack
    @ChepyDecorators.is_stdout
    def password_hashing(self, format: str, **kwargs: Any) -> HashingT:
        """Hash the state with various password hashing algorithms

        Args:
            format (str): Hash state with password algorithms

        Returns:
            Chepy: The Chepy object.
        """
        self.state = getattr(passlib.hash, format).hash(
            self._convert_to_bytes(), **kwargs
        )
        return self



import unicodedata
from typing import TypeVar

import emoji
import regex as re

from ..core import ChepyCore, ChepyDecorators

LanguageT = TypeVar("LanguageT", bound="Language")


class Language(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    @ChepyDecorators.call_stack
    def search_perl_unicode_props(self, lang: str) -> LanguageT:
        """Search using perl unicode properties.
        https://perldoc.perl.org/perluniprops#(%5Cd+)-in-the-info-column-gives-the-number-of-Unicode-code-points-matched-by-this-property.

        Args:
            lang (str): Required. A string value identifying the language.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"\p{" + lang + "}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def find_emojis(self) -> LanguageT:
        """Find emojis, symbols, pictographs, map symbols and flags

        Returns:
            Chepy: The Chepy object.
        """
        self.state = [e.get("emoji") for e in emoji.emoji_list(self._convert_to_str())]
        return self

    @ChepyDecorators.call_stack
    def encode(self, encoding: str, errors: str = "backslashreplace") -> LanguageT:
        """Encode the string using the given encoding.

        Args:
            encoding (str): Encoding to use.
            errors (str, optional): How to handle errors when encoding. Defaults to 'backslashreplace'.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_str().encode(encoding, errors=errors)
        return self

    @ChepyDecorators.call_stack
    def encode_us_ascii_7_bit(self) -> LanguageT:
        """Encode state using US ascii 7 bit

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_str()
        self.state = "".join(chr(ord(c) & 127) for c in data)
        return self

    @ChepyDecorators.call_stack
    def decode(self, encoding: str, errors: str = "backslashreplace") -> LanguageT:
        """Decode the string using the given encoding.

        Args:
            encoding (str): Encoding to use.
            errors (str, optional): How to handle errors when decoding. Defaults to 'backslashreplace'.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_bytes().decode(encoding, errors=errors)
        return self

    @ChepyDecorators.call_stack
    def remove_diacritics(self) -> LanguageT:
        """Replaces accented characters latin character equivalent.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = unicodedata.normalize("NFKD", self._convert_to_str()).encode(
            "ascii", errors="ignore"
        )
        return self

    @ChepyDecorators.call_stack
    def unicode_to_str(self, as_bytes=False) -> LanguageT:
        """Escape any \\u characters to its proper unicode representation

        Args:
            as_bytes (bool): Treat state as bytes. This does not handle %u or U+ encodings

        Returns:
            Chepy: The Chepy object.
        """
        if as_bytes:
            self.state = self._convert_to_bytes().decode(
                "unicode-escape", errors="backslashreplace"
            )
        else:
            data = self._convert_to_str()
            cleaned_string = re.sub(r"(\\u|%u|U\+)", r"\\u", data)
            self.state = bytes(cleaned_string, "utf-8").decode(
                "unicode-escape", errors="backslashreplace"
            )
        return self

    @ChepyDecorators.call_stack
    def str_to_unicode(self, prefix: str = "\\u", all_chars: bool = False) -> LanguageT:
        """Convert unicode to str

        Args:
            prefix (str): Prefix character.
            all_chars (bool): Force convert all chars to unicode.

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_str()
        if all_chars:
            hold = []
            for d in data:
                hold.append("{}{:04x}".format(prefix, ord(d)))
            self.state = "".join(hold)
            return self
        self.state = data.encode("unicode_escape")
        return self



from ..core import ChepyCore
from typing import Any, TypeVar

NetworkingT = TypeVar('NetworkingT', bound='Networking')

class Networking(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def defang_url(self: NetworkingT) -> NetworkingT: ...
    def refang_url(self: NetworkingT) -> NetworkingT: ...
    def defang_ip(self: NetworkingT) -> NetworkingT: ...
    def refang_ip(self: NetworkingT) -> NetworkingT: ...
    def parse_uri(self: NetworkingT) -> NetworkingT: ...
    def parse_ip_range(self: NetworkingT) -> NetworkingT: ...
    def parse_ipv6(self: NetworkingT) -> NetworkingT: ...
    def int_to_ip(self: NetworkingT) -> NetworkingT: ...
    def ip_to_int(self: NetworkingT) -> NetworkingT: ...
    def get_ssl_cert(self: NetworkingT, port: int=...) -> NetworkingT: ...



import logging
from typing import Any, List, Mapping, Tuple, Union, TypeVar, Literal, Callable, Dict

jsonpickle: Any

ChepyCoreT = TypeVar('ChepyCoreT', bound='ChepyCore')

class ChepyDecorators:
    @staticmethod
    def call_stack(func: Any, *args: Any, **kwargs: Any): ...
    @staticmethod
    def is_stdout(func: Any, *args: Any, **kwargs: Any): ...

class ChepyCore:
    states: Any = ...
    buffers: Any = ...
    write: Any = ...
    bake: Any = ...
    cyberchef: Any = ...
    read_file: Any = ...
    log_level: Any = ...
    log_format: str = ...
    _registers: Dict[str, Union[str, bytes]] = ...
    _log: logging.Logger = ...
    def __init__(self, *data: Any) -> None: ...
    def _convert_to_bytes(self) -> bytes: ...
    def _to_bytes(self, data: Any) -> bytes: ...
    def _convert_to_bytearray(self) -> bytearray: ...
    def _convert_to_str(self) -> str: ...
    def _convert_to_int(self) -> int: ...
    def _str_to_bytes(self, s: str) -> bytes: ...
    def _get_nested_value(self: ChepyCoreT, data: dict, key:str, split_by: str=".") -> Any: ...
    def _bytes_to_str(self, s: bytes) -> str: ...
    @property
    def state(self): ...
    @state.setter
    def state(self: ChepyCoreT, val: Any) -> None: ...
    def fork(self: ChepyCoreT, methods: List[Union[Tuple[Union[str, Callable], dict], Tuple[Union[str, Callable[None, ChepyCoreT]],]]]) -> ChepyCoreT: ...
    def for_each(self: ChepyCoreT, methods: List[Union[Tuple[Union[str, Callable], Dict[str, Any]], Tuple[Union[str, Callable[None, ChepyCoreT]],]]], merge: Union[str, bytes, None]=None) -> ChepyCoreT: ...
    def set_state(self: ChepyCoreT, data: Any) -> ChepyCoreT: ...
    def create_state(self: ChepyCoreT): ...
    def copy_state(self: ChepyCoreT, index: int=...) -> ChepyCoreT: ...
    def change_state(self: ChepyCoreT, index: int) -> ChepyCoreT: ...
    def switch_state(self: ChepyCoreT, index: int) -> ChepyCoreT: ...
    def delete_state(self: ChepyCoreT, index: int) -> ChepyCoreT: ...
    def get_state(self: ChepyCoreT, index: int) -> ChepyCoreT: ...
    def save_buffer(self: ChepyCoreT, index: int=...) -> ChepyCoreT: ...
    def load_buffer(self: ChepyCoreT, index: int) -> ChepyCoreT: ...
    def delete_buffer(self: ChepyCoreT, index: int) -> ChepyCoreT: ...
    def substring(self: ChepyCoreT, pattern: str, group: int=...) -> ChepyCoreT: ...
    @property
    def o(self): ...
    @property
    def out(self: ChepyCoreT) -> ChepyCoreT: ...
    def out_as_any(self: ChepyCoreT) -> str: ...
    def get_by_index(self: ChepyCoreT, *indexes: int) -> ChepyCoreT: ...
    def get_by_key(self: ChepyCoreT, *keys: Union[str, bytes], py_style: bool=False, split_key: Union[str, None] = '.') -> ChepyCoreT: ...
    def copy_to_clipboard(self: ChepyCoreT) -> None: ...
    def copy(self: ChepyCoreT) -> None: ...
    def web(self: ChepyCoreT, magic: bool=..., cyberchef_url: str=...) -> None: ...
    def http_request(self: ChepyCoreT, method: str=..., params: dict=..., json: dict=..., headers: dict=..., cookies: dict=...) -> ChepyCoreT: ...
    def load_from_url(self: ChepyCoreT, method: str=..., params: dict=..., json: dict=..., headers: dict=..., cookies: dict=...) -> ChepyCoreT: ...
    def load_dir(self: ChepyCoreT, pattern: str=...) -> ChepyCoreT: ...
    def load_file(self: ChepyCoreT, binary_mode: bool=..., encoding: Union[str, None]=...) -> ChepyCoreT: ...
    def write_to_file(self: ChepyCoreT, path: str) -> None: ...
    def write_binary(self: ChepyCoreT, path: str) -> None: ...
    @property
    def recipe(self) -> List[Dict[str, Union[str, Dict[str, Any]]]]: ...
    def run_recipe(self: ChepyCoreT, recipes: List[Mapping[str, Union[str, Mapping[str, Any]]]]) -> ChepyCoreT: ...
    def save_recipe(self: ChepyCoreT, path: str) -> ChepyCoreT: ...
    def load_recipe(self: ChepyCoreT, path: str) -> ChepyCoreT: ...
    def run_script(self: ChepyCoreT, path: str, save_state: bool=...) -> ChepyCoreT: ...
    def loop(self: ChepyCoreT, iterations: int, callback: Union[str, Callable], args: dict=...) -> ChepyCoreT: ...
    def loop_list(self: ChepyCoreT, callback: Union[str, Callable], args: dict=...) -> ChepyCoreT: ...
    def loop_dict(self: ChepyCoreT, keys: list, callback: Union[str, Callable], args: dict=...) -> ChepyCoreT: ...
    def debug(self: ChepyCoreT, verbose: bool=...) -> ChepyCoreT: ...
    def reset(self: ChepyCoreT) -> ChepyCoreT: ...
    def print(self: ChepyCoreT) -> ChepyCoreT: ...
    def load_command(self: ChepyCoreT) -> ChepyCoreT: ...
    def pretty(self: ChepyCoreT, indent: int=...) -> ChepyCoreT: ...
    def plugins(self: ChepyCoreT, enable: Literal['true', 'false']) -> None: ...
    def set_plugin_path(self: ChepyCoreT, path: str) -> None: ...
    def subsection(self: ChepyCoreT, pattern: str, methods: List[Tuple[Union[str, object], dict]], group: int=...) -> ChepyCoreT: ...
    def callback(self: ChepyCoreT, callback_function: Callable[[Any], Any]) -> ChepyCoreT: ...
    def register(self: ChepyCoreT, pattern: Union[str, bytes], ignore_case: bool=False, multiline: bool=False, dotall: bool=False, unicode: bool=False, extended: bool=False) -> ChepyCoreT: ...
    def prefix(self: ChepyCoreT, data: Union[str, bytes]) -> ChepyCoreT: ...
    def suffix(self: ChepyCoreT, data: Union[str, bytes]) -> ChepyCoreT: ...
    def get_register(self: ChepyCoreT, key: str) -> Union[str, bytes]: ...
    def set_register(self: ChepyCoreT, key: str, val: Union[str, bytes]) -> ChepyCoreT: ...
    def dump_json(self: ChepyCoreT) -> ChepyCoreT: ...
    def walk_dir(self: ChepyCoreT) -> ChepyCoreT: ...
    def search_dir(self: ChepyCoreT, pattern: Union[str, bytes]) -> ChepyCoreT: ...



import math
from binascii import unhexlify
import base64
from typing import TypeVar, Union, List
from urllib.parse import urlparse as _pyurlparse
import lazy_import

import regex as re
import re as old_re

parsel = lazy_import.lazy_module("parsel")

from ..core import ChepyCore, ChepyDecorators

ExtractorsT = TypeVar("ExtractorsT", bound="Extractors")

_zw_chars = []
_zw_codelengthText = 0
_zw_radix = 0


class Extractors(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    def _parsel_obj(self):
        """Returns a parsel.Selector object"""
        return parsel.Selector(self._convert_to_str())

    @ChepyDecorators.call_stack
    def extract_hashes(self) -> ExtractorsT:
        """Extract md5, sha1, sha256 and sha512 hashes

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(
            >>>     ["60b725f10c9c85c70d97880dfe8191b3", "3f786850e387550fdab836ed7e6dc881de23001b"]
            >>> ).extract_hashes()
            {'md5': [b'60b725f10c9c85c70d97880dfe8191b3'], 'sha1': [b'3f786850e387550fdab836ed7e6dc881de23001b'], 'sha256': [], 'sha512': []}
        """
        # TODO make this more efficient. because at the moment, we are compiling and running this 4 separate times.
        data = self._convert_to_bytes()
        found = {}
        found["md5"] = re.findall(
            rb"(?:[^a-fA-F\d]|\b)([a-fA-F\d]{32})(?:[^a-fA-F\d]|\b)", data
        )
        found["sha1"] = re.findall(
            rb"(?:[^a-fA-F\d]|\b)([a-fA-F\d]{40})(?:[^a-fA-F\d]|\b)", data
        )
        found["sha256"] = re.findall(
            rb"(?:[^a-fA-F\d]|\b)([a-fA-F\d]{64})(?:[^a-fA-F\d]|\b)", data
        )
        found["sha512"] = re.findall(
            rb"(?:[^a-fA-F\d]|\b)([a-fA-F\d]{128})(?:[^a-fA-F\d]|\b)", data
        )
        self.state = found
        return self

    @ChepyDecorators.call_stack
    def extract_strings(
        self, length: int = 4, join_by: Union[str, bytes] = "\n"
    ) -> ExtractorsT:
        """Extract strings from state

        Args:
            length (int, optional): Min length of string. Defaults to 4.
            join_by (str, optional): String to join by. Defaults to newline.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("tests/files/hello").load_file().extract_strings().o
            __PAGEZERO'
            __TEXT'
            __text'
            __TEXT'
            __stubs'
            __TEXT'
            ...
        """
        pattern = b"[^\x00-\x1f\x7f-\xff]{" + str(length).encode() + b",}"
        matches = re.findall(pattern, self._convert_to_bytes())
        self.state = self._str_to_bytes(join_by).join([m for m in matches])
        return self

    @ChepyDecorators.call_stack
    def extract_ips(self, is_binary: bool = False) -> ExtractorsT:
        """Extract ipv4 and ipv6 addresses

        Args:
            is_binary (bool, optional): The state is in binary format. It will then first
                extract the strings from it before matching.

        Returns:
            Chepy: The Chepy object.
        """
        pattern = r"((^\s*((([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))\s*$)|(^\s*((([0-9A-Fa-f]{1,4}:){7}([0-9A-Fa-f]{1,4}|:))|(([0-9A-Fa-f]{1,4}:){6}(:[0-9A-Fa-f]{1,4}|((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){5}(((:[0-9A-Fa-f]{1,4}){1,2})|:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){4}(((:[0-9A-Fa-f]{1,4}){1,3})|((:[0-9A-Fa-f]{1,4})?:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){3}(((:[0-9A-Fa-f]{1,4}){1,4})|((:[0-9A-Fa-f]{1,4}){0,2}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){2}(((:[0-9A-Fa-f]{1,4}){1,5})|((:[0-9A-Fa-f]{1,4}){0,3}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){1}(((:[0-9A-Fa-f]{1,4}){1,6})|((:[0-9A-Fa-f]{1,4}){0,4}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(:(((:[0-9A-Fa-f]{1,4}){1,7})|((:[0-9A-Fa-f]{1,4}){0,5}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:)))(%.+)?\s*$))"
        if is_binary:  # pragma: no cover
            matched = list(
                filter(lambda x: re.search(pattern, x), self.extract_strings().o)
            )
        else:
            matched = list(
                filter(
                    lambda x: re.search(pattern.encode(), x),
                    self._convert_to_bytes().split(),
                )
            )
        self.state = matched
        return self

    @ChepyDecorators.call_stack
    def extract_email(self, is_binary: bool = False) -> ExtractorsT:
        """Extract email

        Args:
            is_binary (bool, optional): The state is in binary format. It will then first
                extract the strings from it before matching.

        Returns:
            Chepy: The Chepy object.

        Examples:
            Sometimes, the state is in a binary format, and not readable. In this case
            set the binary flag to True.

            >>> Chepy("tests/files/test.der").load_file().extract_email(is_binary=True).o
        """
        pattern = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
        if is_binary:
            matched = list(
                filter(
                    lambda x: re.search(pattern.encode(), x),
                    self.extract_strings().o.splitlines(),
                )
            )
        else:  # pragma: no cover
            matched = list(
                filter(
                    lambda x: re.search(pattern, x), self._convert_to_bytes().split()
                )
            )
        self.state = matched
        return self

    @ChepyDecorators.call_stack
    def extract_mac_address(self, is_binary: bool = False) -> ExtractorsT:
        """Extract MAC addresses

        Args:
            is_binary (bool, optional): The state is in binary format. It will then first
                extract the strings from it before matching.

        Returns:
            Chepy: The Chepy object.
        """
        pattern = b"^([0-9a-fA-F][0-9a-fA-F]:){5}([0-9a-fA-F][0-9a-fA-F])$"
        if is_binary:  # pragma: no cover
            matched = list(
                filter(lambda x: re.search(pattern, x), self.extract_strings().o)
            )
        else:
            matched = list(
                filter(
                    lambda x: re.search(pattern, x), self._convert_to_bytes().split()
                )
            )
        self.state = matched
        return self

    @ChepyDecorators.call_stack
    def extract_urls(self, is_binary: bool = False) -> ExtractorsT:
        """Extract urls including http, file, ssh and ftp

        Args:
            is_binary (bool, optional): The state is in binary format. It will then first
                extract the strings from it before matching.

        Returns:
            Chepy: The Chepy object.
        """
        pattern = r"(file|ftps?|http[s]?|ssh)://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
        if is_binary:  # pragma: no cover
            matched = list(
                filter(lambda x: re.search(pattern, x), self.extract_strings().o)
            )
        else:
            matched = list(
                filter(
                    lambda x: re.search(pattern.encode(), x),
                    self._convert_to_bytes().split(),
                )
            )
        self.state = matched
        return self

    @ChepyDecorators.call_stack
    def extract_domains(self, is_binary: bool = False) -> ExtractorsT:
        """Extract domains

        Args:
            is_binary (bool, optional): The state is in binary format. It will then first
                extract the strings from it before matching.

        Returns:
            Chepy: The Chepy object.
        """
        if is_binary:  # pragma: no cover
            matched = list(_pyurlparse(x).netloc for x in self.extract_strings().o)
        else:
            matched = list(
                _pyurlparse(x).netloc
                for x in self._convert_to_bytes().split()
                if x.startswith(b"http")
            )
        self.state = matched
        return self

    @ChepyDecorators.call_stack
    def javascript_comments(self) -> ExtractorsT:
        """Extract javascript comments

        Some false positives is expected because of inline // comments

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"/\*[\w'\s\r\n\*]*\*/|//[\w\s']*|/\*.+?\*/", self._convert_to_str()
        )
        return self

    @ChepyDecorators.call_stack
    def extract_google_api(self) -> ExtractorsT:
        """Extract Goolge api keys

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"AIza[0-9A-Za-z-_]{35}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_google_captcha(self) -> ExtractorsT:
        """Extract Goolge captcha keys

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"6L[0-9A-Za-z-_]{38}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_google_oauth(self) -> ExtractorsT:
        """Extract Goolge oauth keys

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"ya29\.[0-9A-Za-z\-_]+", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_aws_keyid(self) -> ExtractorsT:
        """Extract AWS key ids

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"AKIA[0-9A-Z]{16}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_aws_s3_url(self) -> ExtractorsT:
        """Extract AWS S3 URLs

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"s3\.amazonaws.com[/]+|[a-zA-Z0-9_-]*\.s3\.amazonaws.com",
            self._convert_to_str(),
        )
        return self

    @ChepyDecorators.call_stack
    def extract_facebook_access_token(self) -> ExtractorsT:
        """Extract Facebook access tokens

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"EAACEdEose0cBA[0-9A-Za-z]+", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_auth_basic(self) -> ExtractorsT:
        """Extract basic authentication tokens

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"[B|b]asic\s*[a-zA-Z0-9=:_\+\/-]+", self._convert_to_str()
        )
        return self

    @ChepyDecorators.call_stack
    def extract_auth_bearer(self) -> ExtractorsT:
        """Extract bearer authentication tokens

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"[b|B]earer\s*[a-zA-Z0-9_\-\.=:_\+\/]+", self._convert_to_str()
        )
        return self

    @ChepyDecorators.call_stack
    def extract_mailgun_api(self) -> ExtractorsT:
        """Extract Mailgun API key

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"key-[0-9a-zA-Z]{32}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_twilio_api(self) -> ExtractorsT:
        """Extract Twilio API key

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"SK[0-9a-fA-F]{32}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_twilio_sid(self) -> ExtractorsT:
        """Extract Twilio account or app sid

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"A[C|P][a-zA-Z0-9_\-]{32}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_paypal_bt(self) -> ExtractorsT:
        """Extract Paypal braintree access token

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"access_token\$production\$[0-9a-z]{16}\$[0-9a-f]{32}",
            self._convert_to_str(),
        )
        return self

    @ChepyDecorators.call_stack
    def extract_square_oauth(self) -> ExtractorsT:
        """Extract Square oauth secret token

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"sq0csp-[ 0-9A-Za-z\-_]{43}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_square_access(self) -> ExtractorsT:
        """Extract Square access token

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"sqOatp-[0-9A-Za-z\-_]{22}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_stripe_api(self) -> ExtractorsT:
        """Extract Stripe standard or restricted api token

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(r"[s|r]k_live_[0-9a-zA-Z]{24}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def extract_github(self) -> ExtractorsT:
        """Extract Github access token

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"[a-zA-Z0-9_-]*:[a-zA-Z0-9_\-]+@github\.com*", self._convert_to_str()
        )
        return self

    @ChepyDecorators.call_stack
    def extract_rsa_private(self) -> ExtractorsT:
        """Extract RSA private key

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"-----BEGIN RSA PRIVATE KEY-----", self._convert_to_str()
        )
        return self

    @ChepyDecorators.call_stack
    def extract_dsa_private(self) -> ExtractorsT:
        """Extract DSA private key

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"-----BEGIN DSA PRIVATE KEY-----", self._convert_to_str()
        )
        return self

    @ChepyDecorators.call_stack
    def extract_jwt_token(self) -> ExtractorsT:
        """Extract JWT token

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.findall(
            r"ey[A-Za-z0-9_-]*\.[A-Za-z0-9._-]*|ey[A-Za-z0-9_\/+-]*\.[A-Za-z0-9._\/+-]*",
            self._convert_to_str(),
        )
        return self

    @ChepyDecorators.call_stack
    def extract_base64(self, min: int = 20) -> ExtractorsT:
        """Extract base64 encoded strings

        Args:
            min (int, optional): Minimum length to match. Defaults to 20.

        Returns:
            Chepy: The Chepy object.
        """
        found = re.findall("[a-zA-Z0-9+/=]{%s,}" % str(20), self._convert_to_str())
        if len(found) > 1:  # pragma: no cover
            self.state = found
        else:
            self.state = found[0]
        return self

    @ChepyDecorators.call_stack
    def find_continuous_patterns(
        self, str2: Union[str, bytes], min_value: int = 10
    ) -> ExtractorsT:
        """Find continius patterns between the state as a string and the provided str2

        Args:
            str2 (Union[str, bytes]): String to find matches against
            min_value (int, optional): Minimum value of continuous matches. Defaults to 10.

        Returns:
            Chepy: The Chepy object.
        """
        str1 = self._convert_to_bytes()
        if isinstance(str2, str):
            str2 = str2.encode()
        combined_data = str1 + str2
        data_length = len(combined_data)
        patterns = []

        for length in range(1, data_length + 1):
            for start in range(data_length - length + 1):
                pattern = combined_data[start : start + length]

                if pattern in str1 and pattern in str2 and len(pattern) > min_value:
                    patterns.append(pattern)

        self.state = patterns
        return self

    @ChepyDecorators.call_stack
    def find_longest_continious_pattern(self, str2: str) -> ExtractorsT:
        """Find longest continuous pattern

        Args:
            str2 (Union[str, bytes]): String to find match against

        Returns:
            Chepy: The Chepy object.
        """
        str1 = self._convert_to_bytes()
        if isinstance(str2, str):
            str2 = str2.encode()
        combined_data = str1 + str2
        data_length = len(combined_data)
        matches = []

        for length in range(1, data_length + 1):
            for start in range(data_length - length + 1):
                pattern = combined_data[start : start + length]

                if (
                    pattern in str1
                    and pattern in str2
                    and len(pattern) > len(matches[-1:])
                ):
                    matches.append(pattern)

        self.state = max(matches, key=len) if matches else ""
        return self

    @ChepyDecorators.call_stack
    def extract_zero_width_chars_tags(self) -> ExtractorsT:
        """Extract zero width characters between U+E0000 to U+E007F. Implements
        https://www.irongeek.com/i.php?page=security/unicode-steganography-homoglyph-encoder

        Returns:
            Chepy: The Chepy object.
        """
        input_string = self._convert_to_str()
        extracted_characters = []

        for char in input_string:
            if "\U000e0000" <= char <= "\U000e007f":
                extracted_characters.append(char)

        self.state = unhexlify(
            b"".join(
                [bytes(x.encode("unicode_escape"))[-2:] for x in extracted_characters]
            )
        )
        return self

    @ChepyDecorators.call_stack
    def decode_zero_width(
        self, _zw_chars: str = "\u200c\u200d\u202c\ufeff"
    ) -> ExtractorsT:
        """Extract zero with characters. Decode implementation of
        https://330k.github.io/misc_tools/unicode_steganography.html

        Args:
            chars (str, optional): Characters for stego. Defaults to '\u200c\u200d\u202c\ufeff'.

        Returns:
            Chepy: The Chepy object.
        """

        def set_use_chars(newchars):
            global _zw_chars, _zw_radix, _zw_codelengthText
            if len(newchars) >= 2:
                _zw_chars = list(newchars)
                _zw_radix = len(_zw_chars)
                _zw_codelengthText = math.ceil(math.log(65536) / math.log(_zw_radix))
            return None

        def split_zerowidth_characters(str1):
            result = {}
            result["originalText"] = old_re.sub(
                "[" + "".join(_zw_chars) + "]", "", str1
            )
            result["hiddenText"] = old_re.sub("[^" + "".join(_zw_chars) + "]", "", str1)
            return result

        def decode_from_zero_width_characters_text(str1):
            r = str1
            result = []
            for i in range(_zw_radix):
                r = r.replace(_zw_chars[i], str(i))
            for i in range(0, len(r), _zw_codelengthText):
                result.append(chr(int(r[i : i + _zw_codelengthText], _zw_radix)))
            return "".join(result)

        def decodeText(text):
            split = split_zerowidth_characters(text)
            return {
                # 'originalText': split['originalText'],
                "hidden": decode_from_zero_width_characters_text(
                    split["hiddenText"]
                )  # , codelengthText)
            }

        set_use_chars(_zw_chars)
        self.state = decodeText(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def xpath_selector(self, query: str, index: int = None, namespaces: str = None):
        """Extract data using valid xpath selectors

        Args:
            query (str): Required. Xpath query
            index (int, optional): Matched index. Get the matched node by index. Defaults to None.
            namespaces (str, optional): Namespace. Applies for XML data. Defaults to None.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("http://example.com")
            >>> c.http_request()
            >>> c.xpath_selector("//title/text()", index=0)
            >>> c.o
            "Example Domain"
        """
        s = (
            parsel.Selector(self._convert_to_str(), namespaces=namespaces)
            .xpath(query)
            .getall()
        )
        self.state = s if (index is None) else s[int(index)]
        return self

    @ChepyDecorators.call_stack
    def css_selector(self, query: str):
        """Extract data using valid CSS selectors

        Args:
            query (str): Required. CSS query

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("http://example.com")
            >>> c.http_request()
            >>> c.css_selector("title")
            >>> c.get_by_index(0)
            >>> c.o
            "<title>Example Domain</title>"
        """
        self.state = self._parsel_obj().css(query).getall()
        return self

    @ChepyDecorators.call_stack
    def extract_html_tags(self, tags: List[str]):
        """Extract tags from html along with their attributes

        Args:
            tag (str): A HTML tag

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("http://example.com").http_request().html_tags(['p']).o
            [
                {'tag': 'p', 'attributes': {}},
                {'tag': 'p', 'attributes': {}},
                {'tag': 'p', 'attributes': {}}
            ]
        """
        hold = []

        for tag in tags:
            for element in self._parsel_obj().xpath("//{}".format(tag)):
                attributes = []
                for index, attribute in enumerate(element.xpath("@*"), start=1):
                    attribute_name = element.xpath(
                        "name(@*[%d])" % index
                    ).extract_first()
                    attributes.append((attribute_name, attribute.extract()))
                hold.append({"tag": tag, "attributes": dict(attributes)})

        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def extract_html_comments(self):
        """Extract html comments

        Returns:
            Chepy: The Chepy object.
        """
        self.state = list(
            filter(lambda x: x != "", self._parsel_obj().xpath("//comment()").getall())
        )
        return self

    @ChepyDecorators.call_stack
    def aws_account_id_from_access_key(self):
        """Extract AWS account id from access key

        Returns:
            Chepy: The Chepy object.
        """
        trimmed_AWSKeyID = self._convert_to_str()[4:]
        x = base64.b32decode(trimmed_AWSKeyID)
        y = x[0:6]
        z = int.from_bytes(y, byteorder="big", signed=False)
        mask = int.from_bytes(unhexlify(b"7fffffffff80"), byteorder="big", signed=False)

        self.state = (z & mask) >> 7
        return self



import time
from datetime import datetime
from typing import TypeVar

from ..core import ChepyCore, ChepyDecorators

DateTimeT = TypeVar("DateTimeT", bound="DateTime")


class DateTime(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    @ChepyDecorators.call_stack
    def from_unix_timestamp(self, format: str = "%c", utc: bool = False) -> DateTimeT:
        """Convert UNIX timestamp to datetime

        Args:
            format: Format to use for datetime.strftime()
            utc: Whether to use UTC or local timezone

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("1573426649").from_unix_timestamp()
            "Sun Nov 10 17:57:29 2019"
        """
        data = self._convert_to_int()
        if utc:
            self.state = datetime.utcfromtimestamp(data).strftime(format)
        else:
            self.state = datetime.fromtimestamp(data).strftime(format)
        return self

    @ChepyDecorators.call_stack
    def to_unix_timestamp(self) -> DateTimeT:  # pragma: no cover
        """Convert datetime string to unix ts

        The format for the string is %a %b %d %H:%M:%S %Y, which is equivalent to
        %c from datatime.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("Sun Nov 10 17:57:29 2019").to_unix_timestamp()
            "1573426649"
        """
        self.state = int(
            time.mktime(time.strptime(self._convert_to_str(), "%a %b %d %H:%M:%S %Y"))
        )
        return self



from ..core import ChepyCore
from typing import Any, TypeVar, Literal

pydash: Any
CodeTidyT = TypeVar('CodeTidyT', bound='CodeTidy')

class CodeTidy(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def minify_json(self: CodeTidyT) -> CodeTidyT: ...
    def beautify_json(self: CodeTidyT, indent: int=...) -> CodeTidyT: ...
    def to_upper_case(self: CodeTidyT, by: Literal['all', 'word', 'sentence']=...) -> CodeTidyT: ...
    def to_lower_case(self: CodeTidyT) -> CodeTidyT: ...
    def to_snake_case(self: CodeTidyT) -> CodeTidyT: ...
    def to_camel_case(self: CodeTidyT, ignore_space: bool=...) -> CodeTidyT: ...
    def to_kebab_case(self: CodeTidyT) -> CodeTidyT: ...
    def swap_case(self: CodeTidyT) -> CodeTidyT: ...
    def to_leetspeak(self: CodeTidyT, special_chars: bool=...) -> CodeTidyT: ...
    def random_case(self) -> CodeTidyT: ...



from ..core import ChepyCore
from typing import Any, TypeVar, Union, Literal

AritmeticLogicT = TypeVar("AritmeticLogicT", bound="AritmeticLogic")

class AritmeticLogic(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def add(self: AritmeticLogicT, n: Union[int, str]) -> AritmeticLogicT: ...
    def addition(self: AritmeticLogicT, delimiter: Union[str, None]=None) -> AritmeticLogicT: ...
    def sub(self: AritmeticLogicT, n: Union[int, str]) -> AritmeticLogicT: ...
    def subtract(self: AritmeticLogicT, delimiter: Union[str, None]=None) -> AritmeticLogicT: ...
    def multiply(self: AritmeticLogicT, n: int) -> AritmeticLogicT: ...
    def divide(self: AritmeticLogicT, n: int) -> AritmeticLogicT: ...
    def power(self: AritmeticLogicT, n: int) -> AritmeticLogicT: ...
    def sum(self: AritmeticLogicT) -> AritmeticLogicT: ...
    def mean(self: AritmeticLogicT) -> AritmeticLogicT: ...
    def median(self: AritmeticLogicT) -> AritmeticLogicT: ...
    def int_to_base(self: AritmeticLogicT, base: Union[int, str]) -> AritmeticLogicT: ...
    def bit_shift_right(self: AritmeticLogicT, amount: int=1, operation_type: Literal['logical', 'arithmetic']='logical') -> AritmeticLogicT: ...
    def bit_shift_left(self: AritmeticLogicT, amount: int=1) -> AritmeticLogicT: ...



from ..core import ChepyCore
from typing import Any, Literal, TypeVar

LanguageT = TypeVar("LanguageT", bound="Language")

ENCODINGS = Literal[
    "utf_16_le",
    "utf_16_be",
    "utf_7",
    "utf_8",
    "cp500",
    "cp037",
    "cp874",
    "cp932",
    "gbk",
    "gb2312",
    "cp949",
    "cp950",
    "cp1250",
    "cp1251",
    "cp1252",
    "cp1253",
    "cp1254",
    "cp1255",
    "cp1256",
    "cp1257",
    "cp1258",
    "iso8859_2",
    "iso8859_3",
    "iso8859_4",
    "iso8859_5",
    "iso8859_6",
    "iso8859_7",
    "iso8859_8",
    "iso8859_9",
    "iso8859_10",
    "iso8859_11",
    "iso8859_13",
    "iso8859_14",
    "iso8859_15",
    "ascii",
    "unicode-escape",
]

class Language(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def search_perl_unicode_props(self: LanguageT, lang: str) -> LanguageT: ...
    def find_emojis(self: LanguageT) -> LanguageT: ...
    def encode(self: LanguageT, encoding: ENCODINGS, errors: Literal['ignore', 'replace', 'backslashreplace']=...) -> LanguageT: ...
    def encode_us_ascii_7_bit(self: LanguageT) -> LanguageT: ...
    def decode(self: LanguageT, encoding: ENCODINGS, errors: Literal['ignore', 'replace', 'backslashreplace']=...) -> LanguageT: ...
    def remove_diacritics(self: LanguageT) -> LanguageT: ...
    def unicode_to_str(self: LanguageT, as_bytes: bool=False) -> LanguageT: ...
    def str_to_unicode(self: LanguageT, prefix: Literal['\\u', '%u', 'U+']='\\u', all_chars: bool=False) -> LanguageT: ...



from ..core import ChepyCore
from typing import Any, Literal, TypeVar, Union

UtilsT = TypeVar('UtilsT', bound='Utils')

class Utils(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def reverse(self: UtilsT, count: int=...) -> UtilsT: ...
    def count_occurances(self: UtilsT, regex: str, case_sensitive: bool=...) -> UtilsT: ...
    def remove_newlines(self: UtilsT) -> UtilsT: ...
    def remove_whitespace(self: UtilsT, spaces: bool=..., carriage_return: bool=..., line_feeds: bool=..., tabs: bool=..., form_feeds: bool=...) -> UtilsT: ...
    def remove_nullbytes(self: UtilsT) -> UtilsT: ...
    def regex_search(self: UtilsT, pattern: str, is_bytes: bool=False, ignore_case: bool=False, multiline: bool=False, dotall: bool=False, unicode: bool=False, extended: bool=False) -> UtilsT: ...
    def split_by_char(self: UtilsT, delimiter: str=...) -> UtilsT: ...
    def split_by_regex(self: UtilsT, pattern: str=..., trim: Any=...) -> UtilsT: ...
    def split_by_n(self: UtilsT, n: int) -> UtilsT: ...
    def split_lines(self: UtilsT) -> UtilsT: ...
    def split_chunks(self: UtilsT, chunk_size: int) -> UtilsT: ...
    def select_every_n(self: UtilsT, n: int, start: int=...) -> UtilsT: ...
    def unique(self: UtilsT) -> UtilsT: ...
    def sort_list(self: UtilsT, reverse: bool=...) -> UtilsT: ...
    def sort_dict_key(self: UtilsT, reverse: bool=...) -> UtilsT: ...
    def sort_dict_value(self: UtilsT, reverse: Any=...) -> UtilsT: ...
    def filter_list(self: UtilsT, by: Union[str, dict], regex: bool=...) -> UtilsT: ...
    def filter_list_by_length(self: UtilsT, length: int, exact: bool=...) -> UtilsT: ...
    def filter_dict_key(self: UtilsT, by: str) -> UtilsT: ...
    def filter_dict_value(self: UtilsT, by: str) -> UtilsT: ...
    def slice(self: UtilsT, start: int=..., end: int=...) -> UtilsT: ...
    def strip_ansi(self: UtilsT) -> UtilsT: ...
    def strip(self: UtilsT, pattern: str, ignore_case: bool=...) -> UtilsT: ...
    def strip_non_printable(self: UtilsT) -> UtilsT: ...
    def find_replace(self: UtilsT, pattern: Union[bytes, str], repl: Union[bytes, str], ignore_case: bool=...) -> UtilsT: ...
    def escape_string(self: UtilsT) -> UtilsT: ...
    def unescape_string(self: UtilsT) -> UtilsT: ...
    def color_hex_to_rgb(self: UtilsT) -> UtilsT: ...
    def diff(self: UtilsT, state: int=..., buffer: int=..., colors: bool=False, swap: bool=False, only_changes: bool=False) -> UtilsT: ...
    def pad(self: UtilsT, width: int, direction: Literal['left', 'right']=..., char: str=...) -> UtilsT: ...
    def count(self: UtilsT) -> UtilsT: ...
    def set(self: UtilsT) -> UtilsT: ...
    def regex_to_str(self: UtilsT, all_combo: bool=...) -> UtilsT: ...
    def shuffle(self: UtilsT) -> UtilsT: ...
    def drop_bytes(self: UtilsT, start: int, length: int) -> UtilsT: ...
    def without(self: UtilsT, *values: Any) -> UtilsT: ...
    def pick(self: UtilsT, *values: Any) -> UtilsT: ...
    def expand_alpha_range(self: UtilsT, join_by: Union[str, None]=None) -> UtilsT: ...
    def split_and_count(self: UtilsT, pattern: Union[str, bytes], threshold: int=None) -> UtilsT: ...



from typing import Dict, List

from .modules.aritmeticlogic import AritmeticLogic
from .modules.codetidy import CodeTidy
from .modules.compression import Compression
from .modules.dataformat import DataFormat
from .modules.datetimemodule import DateTime
from .modules.encryptionencoding import EncryptionEncoding
from .modules.extractors import Extractors
from .modules.hashing import Hashing
from .modules.language import Language
from .modules.links import Links
from .modules.multimedia import Multimedia
from .modules.networking import Networking
from .modules.other import Other
from .modules.publickey import Publickey
from .modules.search import Search
from .modules.utils import Utils

class Chepy(
    AritmeticLogic,
    CodeTidy,
    Compression,
    DataFormat,
    DateTime,
    EncryptionEncoding,
    Extractors,
    Hashing,
    Language,
    Links,
    Multimedia,
    Networking,
    Other,
    Publickey,
    Search,
    Utils,
): ...

def search_chepy_methods(search: str) -> None: ...
def show_plugins() -> Dict[str, List[str]]: ...



import pretty_errors
from docstring_parser import parse as _doc_parse
from types import FunctionType

from .modules.aritmeticlogic import AritmeticLogic
from .modules.codetidy import CodeTidy
from .modules.compression import Compression
from .modules.dataformat import DataFormat
from .modules.datetimemodule import DateTime
from .modules.encryptionencoding import EncryptionEncoding
from .modules.extractors import Extractors
from .modules.hashing import Hashing
from .modules.language import Language
from .modules.links import Links
from .modules.networking import Networking
from .modules.other import Other
from .modules.publickey import Publickey
from .modules.search import Search
from .modules.utils import Utils
from .modules.internal.colors import cyan

from .config import ChepyConfig

_plugins = ChepyConfig().load_plugins()


class Chepy(
    AritmeticLogic,
    CodeTidy,
    Compression,
    DataFormat,
    DateTime,
    EncryptionEncoding,
    Extractors,
    Hashing,
    Language,
    Links,
    Networking,
    Other,
    Publickey,
    Search,
    Utils,
    *_plugins
):
    """Chepy class that exposes all functionality of Chepy and its plugins."""

    pass


def show_plugins():  # pragma: no cover
    hold = {}
    for p in _plugins:
        hold[p.__name__] = [
            name
            for name, attr in vars(p).items()
            if isinstance(attr, FunctionType) and not name.startswith("_")
        ]
    return hold


def search_chepy_methods(search: str) -> None:  # pragma: no cover
    """Search for Chepy methods

    Args:
        search (str): String to search for
    """
    methods = dir(Chepy)
    for method in methods:
        if search in method and not method.startswith("_"):
            docs = _doc_parse(getattr(Chepy, method).__doc__).short_description
            print(cyan(method), docs)



from typing import TypeVar
from uuid import uuid4

from ..core import ChepyCore, ChepyDecorators

OtherT = TypeVar("OtherT", bound="Other")


class Other(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    @ChepyDecorators.call_stack
    def generate_uuid(self) -> OtherT:
        """Generate v4 UUID

        Generates an RFC 4122 version 4 compliant Universally Unique Identifier
        (UUID), also known as a Globally Unique Identifier (GUID). A version 4
        UUID relies on random numbers

        Returns:
            str: A random UUID

        Examples:
            >>> Chepy('').generate_uuid()
            92644a99-632a-47c1-b169-5a141172924b
        """
        self.state = str(uuid4())
        return self



from ..core import ChepyCore
from typing import Any, TypeVar, Union

SearchT = TypeVar('SearchT', bound='Search')

class Search(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def search(self: SearchT, pattern: Union[str, bytes]) -> SearchT: ...
    def search_list(self: SearchT, pattern: Union[str, bytes]) -> SearchT: ...
    def search_ctf_flags(self: SearchT, prefix: str, postfix: str=...) -> SearchT: ...
    def search_slack_tokens(self: SearchT) -> SearchT: ...
    def search_slack_webhook(self: SearchT) -> SearchT: ...
    def search_private_key(self: SearchT) -> SearchT: ...
    def search_twilio_key(self: SearchT) -> SearchT: ...
    def search_aws_key(self: SearchT) -> SearchT: ...



from typing import Any

class StateNotList(Exception):
    def __init__(self, msg: str = ..., *args: Any, **kwargs: Any) -> None: ...

class StateNotDict(Exception):
    def __init__(self, msg: str = ..., *args: Any, **kwargs: Any) -> None: ...



from ..core import ChepyCore
from typing import Any, TypeVar

LinksT = TypeVar('LinksT', bound='Links')

class Links(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def pastebin_to_raw(self: LinksT) -> LinksT: ...
    def github_to_raw(self: LinksT) -> LinksT: ...
    def google_search_ei_to_epoch(self: LinksT) -> LinksT: ...



from typing import TypeVar

import json
import random

import pydash
import regex as re
from ..core import ChepyCore, ChepyDecorators

CodeTidyT = TypeVar("CodeTidyT", bound="CodeTidy")


class CodeTidy(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    @ChepyDecorators.call_stack
    def minify_json(self) -> CodeTidyT:
        """Minify JSON string

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("/path/to/file.json").load_file()
            >>> print(c.minify_json())
        """
        self.state = json.dumps(
            json.loads(self._convert_to_str()), separators=(",", ":")
        )
        return self

    @ChepyDecorators.call_stack
    def beautify_json(self, indent: int = 2) -> CodeTidyT:
        """Beautify minified JSON

        Args:
            indent (int, optional): Indent level. Defaults to 2.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("/path/to/file.json").load_file()
            >>> print(c.beautify_json(indent=4))
        """
        self.state = json.dumps(json.loads(self._convert_to_str()), indent=indent)
        return self

    @ChepyDecorators.call_stack
    def to_upper_case(self, by: str = "all") -> CodeTidyT:
        """Convert string to uppercase

        Args:
            by (str, optional): Convert all, by word or by sentence. Defaults to 'all'.

        Returns:
            Chepy: The Chepy object.

        Examples:
            Uppercase by word

            >>> Chepy("some String").to_upper_case(by="word").o
            "Some String"

            Uppercase by sentence

            >>> Chepy("some String").to_upper_case(by="sentence").o
            "Some string"

            Uppercase all

            >>> Chepy("some String").to_upper_case(by="all").o
            "SOME STRING"
        """
        assert by in [
            "all",
            "word",
            "sentence",
        ], "Valid options are all, word and sentence"
        if by == "all":
            self.state = self._convert_to_str().upper()
        elif by == "word":
            self.state = self._convert_to_str().title()
        elif by == "sentence":
            self.state = self._convert_to_str().capitalize()
        return self

    @ChepyDecorators.call_stack
    def to_lower_case(self) -> CodeTidyT:
        """Convert string to lowercase

        Converts every character in the input to lower case.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("HelLo WorLd").to_lower_case().o
            "hello world"
        """
        self.state = self._convert_to_str().lower()
        return self

    @ChepyDecorators.call_stack
    def to_snake_case(self) -> CodeTidyT:
        """Convert string to snake case

        Converts the input string to snake case. Snake case is all lower case
        with underscores as word boundaries. e.g. this_is_snake_case.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("helloWorld").to_snake_case().o
            "hello_world"
        """
        s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", self._convert_to_str())
        self.state = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()
        return self

    @ChepyDecorators.call_stack
    def to_camel_case(self, ignore_space: bool = False) -> CodeTidyT:
        """Convert string to camel case

        Converts the input string to camel case. Camel case is all lower case
        except letters after word boundaries which are uppercase. e.g. thisIsCamelCase

        Args:
            ignore_space (bool, optional): Ignore space boundaries. Defaults to False.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some Data_test").to_camel_case().o
            "someDataTest"

            To ignore space, we can set the `ignore_space` to True
            >>> Chepy("some Data_test").to_camel_case(ignore_space=True).o
            "some DataTest"
        """
        if ignore_space:
            r = re.compile(r"_.|\-.")
        else:
            r = re.compile(r"_.|\-.|\s.")
        self.state = r.sub(lambda x: x.group()[1].upper(), self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def to_kebab_case(self) -> CodeTidyT:
        """Convert string to kebab case

        Converts the input string to kebab case. Kebab case is all lower case
        with dashes as word boundaries. e.g. this-is-kebab-case.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("Some data_test").to_kebab_case().o
            "some-data-test"
        """
        self.state = pydash.kebab_case(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def swap_case(self) -> CodeTidyT:
        """Swap case in a string

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("SoMe TeXt").swap_case().o
            "sOmE tExT"
        """
        self.state = pydash.swap_case(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def to_leetspeak(self, special_chars: bool = True) -> CodeTidyT:
        """Convert string to l33t speak

        Args:
            special_chars (bool, optional): Use special chars in conversion. Defaults to True.

        Returns:
            Chepy: The Chepy object

        Examples:
            >>> Chepy("somexValue").to_leetspeak().o
            "50m3%V@1u3"
        """
        chars = {"B": "8", "E": "3", "L": "1", "O": "0", "S": "5", "T": "7", "Z": "2"}
        special = {"A": "@", "C": "(", "I": "!", "X": "%"}
        if special_chars:
            chars = {**chars, **special}
        hold = ""
        for char in list(self._convert_to_str()):
            upper = char.upper()
            char_to_append = chars.get(upper)
            if char_to_append:
                hold += char_to_append
            else:
                hold += char
        self.state = hold
        return self

    @ChepyDecorators.call_stack
    def random_case(self) -> CodeTidyT:
        """Randomly change the case

        Returns:
            Chepy: The Chepy object.
        """
        string = self._convert_to_str()
        string_length = len(string)

        random_indices = random.sample(range(string_length), string_length)
        random_chars = []
        for i in random_indices:
            if random.choice([True, False]):
                random_chars.append(string[i].upper())
            else:
                random_chars.append(string[i].lower())

        string_list = list(string)
        for index, char in zip(random_indices, random_chars):
            if 0 <= index < len(string_list):
                string_list[index] = char
        self.state = "".join(string_list)
        return self



from typing import Any

class ChepyConfig:
    chepy_dir: Any = ...
    chepy_conf: Any = ...
    config: Any = ...
    enable_plugins: Any = ...
    plugin_path: Any = ...
    history_path: Any = ...
    prompt_char: Any = ...
    prompt_colors: Any = ...
    show_rprompt: Any = ...
    prompt_rprompt: Any = ...
    prompt_bottom_toolbar: Any = ...
    prompt_toolbar_version: Any = ...
    prompt_toolbar_states: Any = ...
    prompt_toolbar_buffers: Any = ...
    prompt_toolbar_type: Any = ...
    prompt_toolbar_plugins: Any = ...
    prompt_toolbar_errors: Any = ...
    prompt_cli_method: Any = ...
    prompt_plugin_method: Any = ...
    cli_info_color: Any = ...
    def __init__(self) -> None: ...
    def load_plugins(self): ...



from typing import Dict, Iterator, Union, Any

def factordb(n: int) -> dict: ...
def construct_private_key(n: int, e: int, d: int, format: str=..., passphrase: str=...) -> str: ...
def xor_bruteforce_multi(data: str, min: int=..., max: int=..., errors: str=...) -> Iterator[Dict[str, str]]: ...
def xor_repeating_key(data1:bytes, data2: bytes, min: int =..., max: int = ...) -> Union[bytes, None]: ...
def generate_rsa_keypair(bits: int=..., passphrase: str=...) -> Dict[str, Dict[str, Any]] : ...


from ..core import ChepyCore
from typing import Any, TypeVar, Union, List

parsel: Any

ExtractorsT = TypeVar('ExtractorsT', bound='Extractors')

class Extractors(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def extract_hashes(self: ExtractorsT) -> ExtractorsT: ...
    def extract_strings(self: ExtractorsT, length: int=..., join_by: str=...) -> ExtractorsT: ...
    def extract_ips(self: ExtractorsT, is_binary: bool=...) -> ExtractorsT: ...
    def extract_email(self: ExtractorsT, is_binary: bool=...) -> ExtractorsT: ...
    def extract_mac_address(self: ExtractorsT, is_binary: bool=...) -> ExtractorsT: ...
    def extract_urls(self: ExtractorsT, is_binary: bool=...) -> ExtractorsT: ...
    def extract_domains(self: ExtractorsT, is_binary: bool=...) -> ExtractorsT: ...
    def javascript_comments(self: ExtractorsT) -> ExtractorsT: ...
    def extract_google_api(self: ExtractorsT) -> ExtractorsT: ...
    def extract_google_captcha(self: ExtractorsT) -> ExtractorsT: ...
    def extract_google_oauth(self: ExtractorsT) -> ExtractorsT: ...
    def extract_aws_keyid(self: ExtractorsT) -> ExtractorsT: ...
    def extract_aws_s3_url(self: ExtractorsT) -> ExtractorsT: ...
    def extract_facebook_access_token(self: ExtractorsT) -> ExtractorsT: ...
    def extract_auth_basic(self: ExtractorsT) -> ExtractorsT: ...
    def extract_auth_bearer(self: ExtractorsT) -> ExtractorsT: ...
    def extract_mailgun_api(self: ExtractorsT) -> ExtractorsT: ...
    def extract_twilio_api(self: ExtractorsT) -> ExtractorsT: ...
    def extract_twilio_sid(self: ExtractorsT) -> ExtractorsT: ...
    def extract_paypal_bt(self: ExtractorsT) -> ExtractorsT: ...
    def extract_square_oauth(self: ExtractorsT) -> ExtractorsT: ...
    def extract_square_access(self: ExtractorsT) -> ExtractorsT: ...
    def extract_stripe_api(self: ExtractorsT) -> ExtractorsT: ...
    def extract_github(self: ExtractorsT) -> ExtractorsT: ...
    def extract_rsa_private(self: ExtractorsT) -> ExtractorsT: ...
    def extract_dsa_private(self: ExtractorsT) -> ExtractorsT: ...
    def extract_jwt_token(self: ExtractorsT) -> ExtractorsT: ...
    def extract_base64(self: ExtractorsT, min: int=...) -> ExtractorsT: ...
    def find_continuous_patterns(self: ExtractorsT, str2: Union[str, bytes], min_value: int=...) -> ExtractorsT: ...
    def find_longest_continious_pattern(self: ExtractorsT, str2: Union[str, bytes]) -> ExtractorsT: ...
    def extract_zero_width_chars_tags(self: ExtractorsT) -> ExtractorsT: ...
    def decode_zero_width(self: ExtractorsT, chars: str=...) -> ExtractorsT: ...
    def xpath_selector(self: ExtractorsT, query: str, index:int=None, namespaces: str=None) -> ExtractorsT: ...
    def css_selector(self: ExtractorsT, query: str) -> ExtractorsT: ...
    def extract_html_comments(self: ExtractorsT) -> ExtractorsT: ...
    def extract_html_tags(self: ExtractorsT, tag: List[str]) -> ExtractorsT: ...
    def aws_account_id_from_access_key(self: ExtractorsT) -> ExtractorsT: ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This software is hereby released into public domain. Use it wisely.
#
# Originally written by Mirek Kratochvil (2017)
# Python3 port by Bernhard Esslinger (Feb 2018)

import random

letters = "_abcdefghijklmnopqrstuvwxyz.0123456789,-+*/:?!'()"
tiles = list(zip(letters, map(lambda x: (x // 7, x % 7), range(7 * 7))))


def check_key(key):  # pragma: no cover
    if len(key) != len(letters):
        raise ValueError("Wrong key size")
    cnts = {}
    for c in letters:
        cnts[c] = 0
    for c in key:
        if not c in cnts:
            raise ValueError("Letter " + c + " not in LS47!")
        cnts[c] += 1
        if cnts[c] > 1:
            raise ValueError("Letter " + c + " duplicated in key!")


def find_ix(letter):  # pragma: no cover
    m = [l for l in tiles if l[0] == letter]
    if len(m) != 1:
        raise ValueError("Letter " + letter + " not in LS47!")
    for (l, pos) in m:
        return pos


def find_pos(key, letter):  # pragma: no cover
    p = key.find(letter)
    if p >= 0 and p < 7 * 7:
        return (p // 7, p % 7)
    raise ValueError("Letter " + letter + " not in key?!")


def add_pos(a, b):  # pragma: no cover
    return ((a[0] + b[0]) % 7, (a[1] + b[1]) % 7)


def sub_pos(a, b):
    return ((a[0] - b[0]) % 7, (a[1] - b[1]) % 7)


def find_at_pos(key, coord):  # pragma: no cover
    return key[coord[1] + coord[0] * 7]


def rotate_right(key, row, n):  # pragma: no cover
    mid = key[7 * row : 7 * (row + 1)]
    n = (7 - n % 7) % 7
    return key[: 7 * row] + mid[n:] + mid[:n] + key[7 * (row + 1) :]


def rotate_down(key, col, n):  # pragma: no cover
    lines = [key[i * 7 : (i + 1) * 7] for i in range(7)]
    lefts = [l[:col] for l in lines]
    mids = [l[col] for l in lines]
    rights = [l[col + 1 :] for l in lines]
    n = (7 - n % 7) % 7
    mids = mids[n:] + mids[:n]
    return "".join(lefts[i] + mids[i] + rights[i] for i in range(7))


def rotate_marker_right(m, row, n):  # pragma: no cover
    if m[0] != row:
        return (m[0], m[1])
    else:
        return (m[0], (m[1] + n) % 7)


def rotate_marker_down(m, col, n):  # pragma: no cover
    if m[1] != col:
        return (m[0], m[1])
    else:
        return ((m[0] + n) % 7, m[1])


def derive_key(password):  # pragma: no cover
    i = 0
    k = letters
    for c in password:
        (row, col) = find_ix(c)
        k = rotate_down(rotate_right(k, i, col), i, row)
        i = (i + 1) % 7
    return k


def encrypt(key, plaintext):  # pragma: no cover
    check_key(key)
    mp = (0, 0)
    ciphertext = ""
    for p in plaintext:
        pp = find_pos(key, p)
        mix = find_ix(find_at_pos(key, mp))
        cp = add_pos(pp, mix)
        c = find_at_pos(key, cp)
        ciphertext += c

        key = rotate_right(key, pp[0], 1)
        cp = find_pos(key, c)
        key = rotate_down(key, cp[1], 1)
        mp = add_pos(mp, find_ix(c))
    return ciphertext


def decrypt(key, ciphertext):  # pragma: no cover
    check_key(key)
    mp = (0, 0)
    plaintext = ""
    for c in ciphertext:
        cp = find_pos(key, c)
        mix = find_ix(find_at_pos(key, mp))
        pp = sub_pos(cp, mix)
        p = find_at_pos(key, pp)
        plaintext += p

        key = rotate_right(key, pp[0], 1)
        cp = find_pos(key, c)
        key = rotate_down(key, cp[1], 1)
        mp = add_pos(mp, find_ix(c))
    return plaintext


def encrypt_pad(key, plaintext, signature, padding_size):  # pragma: no cover

    # TODO it would also be great to randomize the message length.

    check_key(key)
    padding = "".join(
        map(lambda x: letters[random.randint(0, len(letters) - 1)], range(padding_size))
    )

    return encrypt(key, padding + plaintext + "---" + signature)


def decrypt_pad(key, padding_size, ciphertext):  # pragma: no cover
    check_key(key)
    return decrypt(key, ciphertext)[padding_size:]



import lazy_import
import random
import difflib
from collections import OrderedDict, Counter
from typing import TypeVar, Union, Any
from .internal.helpers import expand_alpha_range as _ex_al_range

import chepy.modules.internal.colors as _int_colors

exrex = lazy_import.lazy_module("exrex")
import pydash
import regex as re

from ..core import ChepyCore, ChepyDecorators
from .exceptions import StateNotDict, StateNotList

UtilsT = TypeVar("UtilsT", bound="Utils")


class Utils(ChepyCore):
    def __init__(self, *data):
        super().__init__(*data)

    @ChepyDecorators.call_stack
    def reverse(self, count: int = 1) -> UtilsT:
        """Reverses a string

        Args:
            count (int, optional): Reverse by the number of characters indicated in count. Defaults to 1.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("abcdefg").reverse().out
            "gfedcba"
        """
        if count == 1:
            self.state = self.state[::-1]
            return self
        else:
            self.state = "".join(
                [self.state[x : x + count] for x in range(0, len(self.state), count)][
                    ::-1
                ]
            )
            return self

    @ChepyDecorators.call_stack
    def count_occurances(self, regex: str, case_sensitive: bool = False) -> UtilsT:
        """Counts occurrences of the regex.

        Counts the number of times the provided string occurs.

        Args:
            regex (str): Required. Regex string to search for
            case_sensitive (bool, optional): If search should be case insensitive, by default False

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("AABCDADJAKDJHKSDAJSDdaskjdhaskdjhasdkja").count_occurances("ja").out
            2
        """
        if case_sensitive:
            r = re.compile(regex)
        else:
            r = re.compile(regex, re.IGNORECASE)
        self.state = len(r.findall(self._convert_to_str()))
        return self

    @ChepyDecorators.call_stack
    def remove_newlines(self):
        """Remove new lines

        Returns:
            Chepy: The Chepy object.
        """
        self.data = self._convert_to_bytes()
        self.state = self.data.translate(None, b"\r\n")
        return self

    @ChepyDecorators.call_stack
    def remove_whitespace(
        self,
        spaces: bool = True,
        carriage_return: bool = True,
        line_feeds: bool = True,
        tabs: bool = True,
        form_feeds: bool = True,
    ):
        """Remove whitespace from a string

        Args:
            spaces (bool, optional): Remove spaces. Defaults to True.
            carriage_return (bool, optional): Remove carriage return \\r. Defaults to True.
            line_feeds (bool, optional): Remove line feeds \\n. Defaults to True.
            tabs (bool, optional): Temove tabs \\t. Defaults to True.
            form_feeds (bool, optional): Remove form feeds \\f. Defaults to True.

        Returns:
            Chepy: The Chepy object.
        """
        remove = []
        if spaces:
            remove.append(" ")
        if carriage_return:
            remove.append("\r")
        if line_feeds:
            remove.append("\n")
        if tabs:
            remove.append("\t")
        if form_feeds:
            remove.append("\f")
        self.state = re.sub("|".join(remove), "", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def remove_nullbytes(self) -> UtilsT:
        """Remove null \\x00 byes from binary data

        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_bytes().replace(b"\x00", b"")
        return self

    @ChepyDecorators.call_stack
    def regex_search(
        self,
        pattern: str,
        is_bytes: bool = False,
        ignore_case: bool = False,
        multiline: bool = False,
        dotall: bool = False,
        unicode: bool = False,
        extended: bool = False,
    ) -> UtilsT:
        """Regex search on current data. State will be an array of matches.

        Args:
            pattern (str): Required. The regex pattern to search by
            is_bytes (bool, optional): Treat the pattern and state as bytes. Defaults to False.
            ignore_case (bool, optional): Set case insensitive flag. Defaults to False.
            multiline (bool, optional): ^/$ match start/end. Defaults to False.
            dotall (bool, optional): `.` matches newline. Defaults to False.
            unicode (bool, optional): Match unicode characters. Defaults to False.
            extended (bool, optional): Ignore whitespace. Defaults to False.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("loLolololoL")
            >>> c.regex_search("ol", ignore_case=True)
        """
        flags = 0
        if ignore_case:
            flags += re.IGNORECASE
        if multiline:
            flags += re.MULTILINE
        if dotall:
            flags += re.DOTALL
        if unicode:
            flags += re.UNICODE
        if extended:
            flags += re.X
        if is_bytes:
            self.state = re.findall(
                self._to_bytes(pattern), self._convert_to_bytes(), flags=flags
            )
        else:
            self.state = re.findall(pattern, self._convert_to_str(), flags=flags)
        return self

    @ChepyDecorators.call_stack
    def split_by_char(self, delimiter: str = " ") -> UtilsT:
        """Split a string by a delimiter

        Args:
            delimiter (str, optional): Delimiter to split by. Defaults to " ".

        Returns:
            UtilsT: The Chepy object.
        """
        delimiter = self._str_to_bytes(delimiter)
        self.state = self._convert_to_bytes().split(delimiter)
        return self

    @ChepyDecorators.call_stack
    def split_by_regex(self, pattern: str = "\n", trim=True) -> UtilsT:
        """Split a string by the given regex pattern

        Args:
            pattern (str, optional): Pattern to split by. Defaults to '\\n'.
            time (bool, optional): Trim whitespace after split. Defaults to True

        Returns:
            Chepy: The Chepy object.
        """
        if trim:
            self.state = list(
                map(pydash.trim, re.split(pattern, self._convert_to_str()))
            )
        else:
            self.state = re.split(pattern, self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def split_by_n(self, n: int) -> UtilsT:
        """Split a string by n characters.

        Args:
            n (int): n from 0

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some string").split_by_n(2).o[2]
            " s"
        """
        self.state = re.findall(".{1," + str(n) + "}", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def split_lines(self):
        """Split a string by newline characters.

        Returns:
            Chepy: The Chepy object.
        """
        self.state = self._convert_to_str().split()
        return self

    @ChepyDecorators.call_stack
    def select_every_n(self, n: int, start: int = 0) -> UtilsT:
        """Select every nth item from a list or string.

        Index starts at 0

        Args:
            n (int): n from 0
            start (int): starting position. Defaults to 0.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(["a", 1, "lol", "b", True]).select_every_n(3)
            ["a", "b"]
        """
        self.state = self.state[start::n]
        return self

    @ChepyDecorators.call_stack
    def split_chunks(self, chunk_size) -> UtilsT:
        """Split data in chunks

        Args:
            chunk_size (int): Chunk size

        Returns:
            Chepy: The Chepy object.
        """
        data = self.state
        if not isinstance(data, list):
            data = self._convert_to_bytes()
        data_chunks = []
        for i in range(0, len(data), chunk_size):
            data_chunks.append(data[i : i + chunk_size])
        self.state = data_chunks
        return self

    @ChepyDecorators.call_stack
    def unique(self) -> UtilsT:
        """Get an array of unique list items

        Raises:
            StateNotList: If state is not a list

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, list), StateNotList()
        self.state = pydash.uniq(self.state)
        return self

    @ChepyDecorators.call_stack
    def sort_list(self, reverse: bool = False) -> UtilsT:
        """Sort a list

        Args:
            reverse (bool, optional): In reverse order. Defaults to False.

        Raises:
            StateNotList: If state is not list

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy(["a", "b", "1", "2"]).sort_list().o
            ["1", "2", "a", "b"]
        """
        assert isinstance(self.state, list), StateNotList()
        self.state = sorted(
            self.state, key=lambda v: (isinstance(v, str), v), reverse=reverse
        )
        return self

    @ChepyDecorators.call_stack
    def sort_dict_key(self, reverse: bool = False) -> UtilsT:
        """Sort a dictionary by key

        Args:
            reverse (bool, optional): Reverse sort order. Defaults to False.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy({'z': 'string', 'a': True, 'zz': 1, 'aaa': {'bb': 'data'}, 'ccc': [1,'a']})
            >>> c.sort_dict_key(reverse=True)
            {'zz': 1, 'z': 'string', 'ccc': [1, 'a'], 'aaa': {'bb': 'data'}, 'a': True}
        """
        assert isinstance(self.state, dict), StateNotDict()
        self.state = dict(OrderedDict(sorted(self.state.items(), reverse=reverse)))
        return self

    @ChepyDecorators.call_stack
    def sort_dict_value(self, reverse=False) -> UtilsT:
        """Sort dictionary by value

        Args:
            reverse (bool, optional): Reverse sort order. Defaults to False.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy({'z': 'string', 'a': 'True', 'zz': '1', 'aaa': {'bb': 'data'}, 'ccc': [1,'a']})
            >>> c.sort_dict_value()
            {'zz': '1', 'a': 'True', 'ccc': [1, 'a'], 'z': 'string', 'aaa': {'bb': 'data'}}
        """
        assert isinstance(self.state, dict), StateNotDict()
        try:
            self.state = dict(
                OrderedDict(
                    sorted(self.state.items(), reverse=reverse, key=lambda x: x[1])
                )
            )
        except:
            self.state = dict(
                OrderedDict(
                    sorted(self.state.items(), reverse=reverse, key=lambda x: str(x[1]))
                )
            )
        return self

    @ChepyDecorators.call_stack
    def filter_list(self, by: Union[str, dict], regex: bool = True) -> UtilsT:
        """Filter a list by a string regex or dict key

        Args:
            by (Union[str, dict]): If string, supports regex. Or dictionary
            regex (bool, optional): If pattern is a regex. Defaults to True

        Raises:
            StateNotList: If state is not a list

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy('[{"a": 1}, {"b": 2}, {"a": 1, "b": 3}]').str_list_to_list().filter_list("b").o
            [{"b": 2}, {"a": 1, "b": 3}]
        """
        assert isinstance(self.state, list), StateNotList()
        if regex:
            pattern = by if isinstance(self.state[0], str) else by.encode()
            self.state = [f for f in self.state if re.search(pattern, f)]
        else:
            self.state = pydash.filter_(self.state, by)
        if len(self.state) == 1:
            self.state = self.state[0]
        return self

    @ChepyDecorators.call_stack
    def filter_list_by_length(self, length: int, exact: bool = False) -> UtilsT:
        """Filter a list by length by specifying minimum length.

        It will also return items that exceed the specified length.

        Args:
            length (int): Minimum length to match
            exact (bool): Match exact length

        Returns:
            Chepy: The Chepy object.
        """
        assert isinstance(self.state, list), StateNotList()
        if exact:
            self.state = [x for x in self.state if len(str(x)) == int(length)]
        else:
            self.state = [x for x in self.state if len(str(x)) >= int(length)]
        return self

    @ChepyDecorators.call_stack
    def filter_dict_key(self, by: str) -> UtilsT:
        """Filter dictionary by key

        Args:
            by (str): Required. Key to filter by.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy({'some': 'dict', 'another': 'val'}).filter_dict_key('ano')
            {'another': 'val'}
        """
        assert isinstance(self.state, dict), StateNotDict()
        self.state = {
            key: val for (key, val) in self.state.items() if re.search(by, str(key))
        }
        return self

    @ChepyDecorators.call_stack
    def filter_dict_value(self, by: str) -> UtilsT:
        """Filter dictionary by value.

        This method does descend into nested dictionary values.

        Args:
            by (str): Required. Value to filter by.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy({'some': 'dict', 'another': 'val'}).filter_dict_value('val')
            {'another': 'val'}
        """
        assert isinstance(self.state, dict), StateNotDict()
        self.state = {
            key: val for (key, val) in self.state.items() if re.search(by, str(val))
        }
        return self

    @ChepyDecorators.call_stack
    def slice(self, start: int = 0, end: int = None) -> UtilsT:
        """Returns the specified slice

        Args:
            start (int, optional): Start position. Defaults to 0.
            end (int, optional): End position. Defaults to None.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some data").slice(3, 6).o
            "e d"
        """
        self.state = self.state[start:end]
        return self

    @ChepyDecorators.call_stack
    def strip_ansi(self) -> UtilsT:
        """Strip ANSI escape codes from string

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("\033[31mThis is a string\033[0m").strip_ansi().o
            "This is a string"
        """
        self.state = re.sub(
            "[\u001b\u009b][[\\]()#;?]*(?:(?:(?:[a-zA-Z\\d]*(?:;[a-zA-Z\\d]*)*)?\u0007)|(?:(?:\\d{1,4}(?:;\\d{0,4})*)?[\\dA-PRZcf-ntqry=><~]))",
            "",
            self._convert_to_str(),
        )
        return self

    @ChepyDecorators.call_stack
    def strip(self, pattern: str, ignore_case=True) -> UtilsT:
        """Strip matched pattern

        Args:
            pattern (str): Required. Pattern to search
            ignore_case (bool, optional): Case insensitive. Defaults to True.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some some data").strip("some\\s").o
            "data"
        """
        flags = 0
        if ignore_case:
            flags = re.IGNORECASE
        self.state = re.sub(pattern, "", self._convert_to_str(), flags=flags)
        return self

    @ChepyDecorators.call_stack
    def strip_non_printable(self) -> UtilsT:
        """String non printable characters

        Returns:
            Chepy: The Chepy object.
        """
        data = self._convert_to_bytes()
        printable_set = set(range(32, 127)) | {9, 10, 13, 12}
        # Filter out non-printable characters
        self.state = bytes(b for b in data if b in printable_set)
        return self

    @ChepyDecorators.call_stack
    def find_replace(
        self, pattern: Union[bytes, str], repl: Union[bytes, str], ignore_case=True
    ) -> UtilsT:
        """Replace matched pattern with repln

        Args:
            pattern (str): Required. Pattern to search
            repl (str): Required. Pattern to match
            ignore_case (bool, optional): Case insensitive. Defaults to True.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("some some data").find_replace("some\\s", "data").o
            "datadatadata"
        """
        flags = 0
        if ignore_case:
            flags = re.IGNORECASE
        self.state = re.sub(
            self._to_bytes(pattern),
            self._to_bytes(repl),
            self._convert_to_bytes(),
            flags=flags,
        )
        return self

    @ChepyDecorators.call_stack
    def escape_string(self) -> UtilsT:
        """Escape all special characters in a string

        Returns:
            Chepy: The Chepy object.
        """
        self.state = re.escape(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def unescape_string(self) -> UtilsT:
        """Unescape \\ from a string

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("\\$ome' d@ta").unescape_string().o
            "$ome' d@ta"
        """
        self.state = re.sub(r"\\", "", self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def color_hex_to_rgb(self) -> UtilsT:
        """Convert hex color to rgb

        Returns:
            Chepy: The Chepy object.
        """
        self.state = tuple(
            int(self._convert_to_str().strip("#")[i : i + 2], 16) for i in (0, 2, 4)
        )
        return self

    @ChepyDecorators.call_stack
    def diff(
        self,
        state: int = None,
        buffer: int = None,
        colors: bool = False,
        swap: bool = False,
        only_changes: bool = False,
    ):
        """Diff state with another state or buffer

        Args:
            state (int, optional): Index of state to compare against. Defaults to None.
            buffer (int, optional): Index of buffer to compare against. Defaults to None.
            colors (bool, optional): Show colored diff. Defaults to False.
            swap (bool, optional): Swap the diff order. Defaults to False.
            only_changes (bool, optional): Return only changes. Defaults to False.

        Raises:
            TypeError: If both state and buffer is set to True.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("first string", "First $trin")
            >>> # there are two states
            >>> c.diff(state=1) # this will diff state 0 with state 1
            {F->f}irst {-$}strin{+g}
            >>> # optionally set colors=True in the diff method to see colored output
        """
        if state is not None and buffer is None:
            data = self.states.get(state)
            if isinstance(data, bytes):  # pragma: no cover
                data = data.decode()
        elif state is None and buffer is not None:
            data = self.buffers.get(buffer)
            if isinstance(data, bytes):  # pragma: no cover
                data = data.decode()
        else:  # pragma: no cover
            raise TypeError("Only select a state or a buffer to diff against")

        if swap:  # pragma: no cover
            matcher = difflib.SequenceMatcher(None, self._convert_to_str(), data)
        else:
            matcher = difflib.SequenceMatcher(None, data, self._convert_to_str())

        def process_tag(tag, i1, i2, j1, j2) -> UtilsT:  # pragma: no cover
            if tag == "replace":
                if colors:
                    return _int_colors.blue(matcher.b[j1:j2])
                else:
                    return "{" + matcher.a[i1:i2] + "->" + matcher.b[j1:j2] + "}"
            if tag == "delete":
                if colors:
                    return _int_colors.red(matcher.a[i1:i2])
                else:
                    return "{-" + matcher.a[i1:i2] + "}"
            if tag == "equal":
                if only_changes:
                    return ""
                return matcher.a[i1:i2]
            if tag == "insert":
                if colors:
                    return _int_colors.green(matcher.b[j1:j2])
                else:
                    return "{+" + matcher.b[j1:j2] + "}"
            assert False, "Unknown tag %r" % tag

        self.state = "".join(process_tag(*t) for t in matcher.get_opcodes())
        return self

    @ChepyDecorators.call_stack
    def pad(self, width: int, direction: str = "left", char: str = " ") -> UtilsT:
        """Pad string with a character

        Args:
            width (int): Required. Total length of string. The padding is calculated from
                the length of state minus width.
            direction (str, optional): Padding direction. left or right. Defaults to 'left'.
            char (str, optional): Char to fill with. Defaults to ' '.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> c = Chepy("lol").pad(5, char="a")
            lol # this is because "lol" is len 3, and padding is 5 - 2, which is 2, so no
            padding is added
            >>> c = Chepy("lol").pad(8, char="a")
            lolaa # 8 - 3 so the output is padded for 5 chars
            >>> c = Chepy("lol").pad(8, direction="right", char="a")
            aalol
        """
        assert direction in [
            "left",
            "right",
        ], "Direction has to be either left or right"
        if direction == "left":
            self.state = self._convert_to_str().ljust(
                width - len(self._convert_to_str()), char
            )
        elif direction == "right":
            self.state = self._convert_to_str().rjust(
                width - len(self._convert_to_str()), char
            )
        return self

    @ChepyDecorators.call_stack
    def count(self) -> UtilsT:
        """Count anything

        Returns:
            Chepy: The Chepy object.
        """
        self.state = pydash.count_by(self.state)
        return self

    @ChepyDecorators.call_stack
    def set(self) -> UtilsT:
        """Get an array of unique values

        Returns:
            Chepy: The Chepy object.
        """
        self.state = list(set(self.state))
        return self

    @ChepyDecorators.call_stack
    def regex_to_str(self, all_combo: bool = False) -> UtilsT:
        """Convert a regex to a matching string

        Args:
            all_combo (bool, optional): Generate all combos that match regex. Defaults to False.

        Returns:
            Chepy: The Chepy object.
        """
        if all_combo:
            self.state = list(exrex.generate(self._convert_to_str()))
        else:
            self.state = exrex.getone(self._convert_to_str())
        return self

    @ChepyDecorators.call_stack
    def shuffle(self) -> UtilsT:
        """Shuffle the state if it is a list, string or bytes. The state is
        unchanged if any other types.

        Returns:
            Chepy: The Chepy object
        """
        data = self.state
        if not isinstance(
            data,
            (
                bytes,
                str,
                list,
            ),
        ):
            return self
        if isinstance(
            data,
            (
                bytes,
                str,
            ),
        ):
            data = list(data)
        random.shuffle(data)
        self.state = data
        return self

    @ChepyDecorators.call_stack
    def drop_bytes(self, start: int, length: int) -> UtilsT:
        """Drop bytes from starting index up to length

        Args:
            start (int): Starting index
            length (int): Number of bytes to drop

        Raises:
            ValueError: If start or length < -1

        Returns:
            Chepy: The Chepy object
        """
        if start < 0 or length < 0:
            raise ValueError(
                "Start and length must be non-negative integers."
            )  # pragma: no cover

        end = start + length
        data = self._convert_to_bytes()
        self.state = data[:start] + data[end:]
        return self

    @ChepyDecorators.call_stack
    def without(self, *values: Any):
        """Remove specified values from the state. Works on strings, bytes, lists and dicts

        Raises:
            TypeError: If state does not contain valid data

        Returns:
            Chepy: The chepy object.
        """
        collection = self.state
        if isinstance(collection, list):
            self.state = [item for item in collection if item not in values]
        elif isinstance(collection, dict):
            self.state = {
                k: v
                for k, v in collection.items()
                if k not in values and v not in values
            }
        elif isinstance(
            collection,
            (
                bytes,
                str,
            ),
        ):
            if isinstance(collection, str):
                collection = collection.encode()
            byte_values = set()
            for value in values:
                if isinstance(value, str):
                    byte_values.update(value.encode())
                elif isinstance(value, bytes):
                    byte_values.update(value)
            self.state = bytes([char for char in collection if char not in byte_values])
        else:  # pragma: no cover
            raise TypeError("Input should be a list, dictionary, string, or bytes")
        return self

    @ChepyDecorators.call_stack
    def pick(self, *values: Any):
        """Only pick specified values from the state. Works on strings, bytes, lists and dicts

        Raises:
            TypeError: If state does not contain valid data

        Returns:
            Chepy: The chepy object.
        """
        collection = self.state
        if isinstance(collection, list):
            self.state = [item for item in collection if item in values]
        elif isinstance(collection, dict):
            self.state = {
                k: v for k, v in collection.items() if k in values or v in values
            }
        elif isinstance(collection, (bytes, str)):
            if isinstance(collection, str):
                collection = collection.encode()
            byte_values = set()
            for value in values:
                if isinstance(value, str):
                    byte_values.update(value.encode())
                elif isinstance(value, bytes):
                    byte_values.update(value)
            self.state = bytes([char for char in collection if char in byte_values])
        else:  # pragma: no cover
            raise TypeError("Input should be a list, dictionary, string, or bytes")
        return self

    @ChepyDecorators.call_stack
    def expand_alpha_range(self, join_by: Union[str, None] = None):
        """Get all alphanumberic or hex chars for the specified range

        Args:
            join_by (str, optional): Join by. Defaults to Union[str, None].

        Returns:
            Chepy: The Chepy object.
        """
        alph_str = self._convert_to_str()
        self.state = _ex_al_range(alph_str=alph_str, join_by=join_by)
        return self

    @ChepyDecorators.call_stack
    def split_and_count(self, pattern: Union[str, bytes], threshold=None) -> UtilsT:
        """
        Splits a string by a regex pattern, counts occurrences of unique items,
        and returns a dictionary with keys ordered by the number of matches.
        Items with counts below the threshold are excluded.
        Returns a dictionary where keys are unique items from the split,
        values are their counts, and keys are ordered by the count.

        Args:
            pattern (Union[str, bytes]): The regex pattern to split the string.
            threshold (int, optional): Minimum count for an item to be included in the result.
                                        Defaults to None, meaning no threshold is applied.

        Returns:
            Chepy: The Chepy object.

        Examples:
            >>> Chepy("apple,banana,apple,orange,banana,apple").split_and_count(",", 2)
            {b"apple": 3, b"banana": 2}
        """
        data = self._convert_to_bytes()
        pattern = self._to_bytes(pattern)
        # Split the string using the regex pattern
        split_items = re.split(pattern, data)

        # Filter out any empty strings from the split results
        filtered_items = [item for item in split_items if item]

        # Count occurrences of unique items
        item_counts = Counter(filtered_items)

        # Filter items based on the threshold
        if threshold is not None:
            item_counts = {
                item: count
                for item, count in item_counts.items()
                if count >= int(threshold)
            }

        # Sort items by count (in descending order), preserving insertion order for equal counts
        sorted_counts = dict(
            sorted(
                item_counts.items(), key=lambda x: (-x[1], filtered_items.index(x[0]))
            )
        )

        self.state = sorted_counts
        return self



# references: https://github.com/pifantastic/python-rison
import re
import urllib
# pragma: no cover

WHITESPACE = ""

IDCHAR_PUNCTUATION = "_-./~"

NOT_IDCHAR = "".join(
    [
        c
        for c in (chr(i) for i in range(127))
        if not (c.isalnum() or c in IDCHAR_PUNCTUATION)
    ]
)

# Additionally, we need to distinguish ids and numbers by first char.
NOT_IDSTART = "-0123456789"

# Regexp string matching a valid id.
IDRX = "[^" + NOT_IDSTART + NOT_IDCHAR + "][^" + NOT_IDCHAR + "]*"

# Regexp to check for valid rison ids.
ID_OK_RE = re.compile("^" + IDRX + "$", re.M)

# Regexp to find the end of an id when parsing.
NEXT_ID_RE = re.compile(IDRX, re.M)


class ParserException(Exception):  # pragma: no cover
    pass


class Parser(object):  # pragma: no cover
    def __init__(self):
        self.string = None
        self.index = 0

    """
    This parser supports RISON, RISON-A and RISON-O.
    """

    def parse(self, string, format=str):
        if format in [list, "A"]:
            self.string = "!({0})".format(string)
        elif format in [dict, "O"]:
            self.string = "({0})".format(string)
        elif format is str:
            self.string = string
        else:
            raise ValueError(
                """Parse format should be one of str, list, dict,
                'A' (alias for list), '0' (alias for dict)."""
            )

        self.index = 0

        value = self.read_value()
        if self.next():
            raise ParserException("unable to parse rison string %r" % (string,))
        return value

    def read_value(self):
        c = self.next()

        if c == "!":
            return self.parse_bang()
        if c == "(":
            return self.parse_open_paren()
        if c == "'":
            return self.parse_single_quote()
        if c in "-0123456789":
            return self.parse_number()

        # fell through table, parse as an id
        s = self.string
        i = self.index - 1

        m = NEXT_ID_RE.match(s, i)
        if m:
            _id = m.group(0)
            self.index = i + len(_id)
            return _id

        if c:
            raise ParserException("invalid character: '" + c + "'")
        raise ParserException("empty expression")

    def parse_array(self):
        ar = []
        while 1:
            c = self.next()
            if c == ")":
                return ar

            if c is None:
                raise ParserException("unmatched '!('")

            if len(ar):
                if c != ",":
                    raise ParserException("missing ','")
            elif c == ",":
                raise ParserException("extra ','")
            else:
                self.index -= 1
            n = self.read_value()
            ar.append(n)

        return ar

    def parse_bang(self):
        s = self.string
        c = s[self.index]
        self.index += 1
        if c is None:
            raise ParserException('"!" at end of input')
        if c not in self.bangs:
            raise ParserException('unknown literal: "!' + c + '"')
        x = self.bangs[c]
        if callable(x):
            return x(self)

        return x

    def parse_open_paren(self):
        count = 0
        o = {}

        while 1:
            c = self.next()
            if c == ")":
                return o
            if count:
                if c != ",":
                    raise ParserException("missing ','")
            elif c == ",":
                raise ParserException("extra ','")
            else:
                self.index -= 1
            k = self.read_value()

            if self.next() != ":":
                raise ParserException("missing ':'")
            v = self.read_value()

            o[k] = v
            count += 1

    def parse_single_quote(self):
        s = self.string
        i = self.index
        start = i
        segments = []

        while 1:
            if i >= len(s):
                raise ParserException('unmatched "\'"')

            c = s[i]
            i += 1
            if c == "'":
                break

            if c == "!":
                if start < i - 1:
                    segments.append(s[start : i - 1])
                c = s[i]
                i += 1
                if c in "!'":
                    segments.append(c)
                else:
                    raise ParserException('invalid string escape: "!' + c + '"')

                start = i

        if start < i - 1:
            segments.append(s[start : i - 1])
        self.index = i
        return "".join(segments)

    # Also any number start (digit or '-')
    def parse_number(self):
        s = self.string
        i = self.index
        start = i - 1
        state = "int"
        permitted_signs = "-"
        transitions = {"int+.": "frac", "int+e": "exp", "frac+e": "exp"}

        while 1:
            if i >= len(s):
                i += 1
                break

            c = s[i]
            i += 1

            if "0" <= c <= "9":
                continue

            if permitted_signs.find(c) >= 0:
                permitted_signs = ""
                continue

            state = transitions.get(state + "+" + c.lower(), None)
            if state is None:
                break
            if state == "exp":
                permitted_signs = "-"

        self.index = i - 1
        s = s[start : self.index]
        if s == "-":
            raise ParserException("invalid number")
        if re.search("[.e]", s):
            return float(s)
        return int(s)

    # return the next non-whitespace character, or undefined
    def next(self):
        s = self.string
        i = self.index
        c = None

        while 1:
            if i == len(s):
                return None
            c = s[i]
            i += 1
            if c not in WHITESPACE:
                break

        self.index = i
        return c

    bangs = {"t": True, "f": False, "n": None, "(": parse_array}


def loads(s, format=str):
    return Parser().parse(s, format=format)


RE_QUOTE = re.compile("^[-A-Za-z0-9~!*()_.',:@$/]*$")


def quote(x):  # pragma: no cover
    if RE_QUOTE.match(x):
        return x

    return (
        urllib.quote(x.encode("utf-8"))
        .replace("%2C", ",", "g")
        .replace("%3A", ":", "g")
        .replace("%40", "@", "g")
        .replace("%24", "$", "g")
        .replace("%2F", "/", "g")
        .replace("%20", "+", "g")
    )


class Encoder(object):  # pragma: no cover
    def __init__(self):
        pass

    @staticmethod
    def encoder(v):
        if isinstance(v, list):
            return Encoder.list
        elif isinstance(v, (str)):
            return Encoder.string
        elif isinstance(v, bool):
            return Encoder.bool
        elif isinstance(v, (float, int)):
            return Encoder.number
        elif isinstance(v, type(None)):
            return Encoder.none
        elif isinstance(v, dict):
            return Encoder.dict
        else:
            raise AssertionError("Unable to encode type: {0}".format(type(v)))

    @staticmethod
    def encode(v):
        encoder = Encoder.encoder(v)
        return encoder(v)

    @staticmethod
    def list(x):
        a = ["!("]
        b = None
        for i in range(len(x)):
            v = x[i]
            f = Encoder.encoder(v)
            if f:
                v = f(v)
                if isinstance(v, (str)):
                    if b:
                        a.append(",")
                    a.append(v)
                    b = True
        a.append(")")
        return "".join(a)

    @staticmethod
    def number(v):
        return str(v).replace("+", "")

    @staticmethod
    def none(_):
        return "!n"

    @staticmethod
    def bool(v):
        return "!t" if v else "!f"

    @staticmethod
    def string(v):
        if v == "":
            return "''"

        if ID_OK_RE.match(v):
            return v

        def replace(match):
            if match.group(0) in ["'", "!"]:
                return "!" + match.group(0)
            return match.group(0)

        v = re.sub(r"([\'!])", replace, v)

        return "'" + v + "'"

    @staticmethod
    def dict(x):
        a = ["("]
        b = None
        ks = sorted(x.keys())
        for i in ks:
            v = x[i]
            f = Encoder.encoder(v)
            if f:
                v = f(v)
                if isinstance(v, (str)):
                    if b:
                        a.append(",")
                    a.append(Encoder.string(i))
                    a.append(":")
                    a.append(v)
                    b = True

        a.append(")")
        return "".join(a)


def encode_array(v):  # pragma: no cover
    if not isinstance(v, list):
        raise AssertionError("encode_array expects a list argument")
    r = dumps(v)
    return r[2 : len(r) - 1]


def encode_object(v):  # pragma: no cover
    if not isinstance(v, dict) or v is None or isinstance(v, list):
        raise AssertionError("encode_object expects an dict argument")
    r = dumps(v)
    return r[1 : len(r) - 1]


def encode_uri(v):  # pragma: no cover
    return quote(dumps(v))


def dumps(string):
    return Encoder.encode(string)
